"""
Distributed energy storage model.
"""

from andes.core.param import NumParam

from ams.core.model import Model

from ams.models.distributed.pvd1 import PVD1Data


class ESD1Data(PVD1Data):
    """
    ESD1 model data.
    """

    def __init__(self):
        PVD1Data.__init__(self)

        self.SOCmin = NumParam(default=0.0, tex_name='SOC_{min}',
                               info='Minimum required value for SOC in limiter',
                               )

        self.SOCmax = NumParam(default=1.0, tex_name='SOC_{max}',
                               info='Maximum allowed value for SOC in limiter',
                               )

        self.SOCinit = NumParam(default=0.5, tex_name='SOC_{init}',
                                info='Initial state of charge',
                                )

        self.En = NumParam(default=100.0, tex_name='E_n',
                           info='Rated energy capacity',
                           unit="MWh"
                           )

        self.EtaC = NumParam(default=1.0, tex_name='Eta_C',
                             info='Efficiency during charging',
                             vrange=(0, 1),
                             )

        self.EtaD = NumParam(default=1.0, tex_name='Eta_D',
                             info='Efficiency during discharging',
                             vrange=(0, 1),
                             )


class ESD1(ESD1Data, Model):
    """
    Distributed energy storage model, revised from ANDES ``ESD1`` model for
    scheduling.

    Following parameters are omitted from the original dynamic model:
    ``fn``, ``busf``, ``xc``, ``pqflag``, ``igreg``, ``v0``, ``v1``,
    ``dqdv``, ``fdbd``, ``ddn``, ``ialim``, ``vt0``, ``vt1``, ``vt2``,
    ``vt3``, ``vrflag``, ``ft0``, ``ft1``, ``ft2``, ``ft3``, ``frflag``,
    ``tip``, ``tiq``, ``recflag``.

    Reference:

    [1] ANDES Documentation, ESD1

    Available:

    https://docs.andes.app/en/latest/groupdoc/DG.html#esd1
    """

    def __init__(self, system, config):
        ESD1Data.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'DG'



"""Callback functions for CPF
"""

from numpy import c_, r_, amax, argmax, array
from scipy.sparse import issparse

import sys
import os


def cpf_default_callback(k, V_c, lam_c, V_p, lam_p, cb_data, cb_state, cb_args, results=None, is_final=False):
    """Default callback function for CPF
    """

    # initialize plotting options
    plot_level = cb_data["ppopt"]["CPF_PLOT_LEVEL"]
    plot_bus = cb_data["ppopt"]["CPF_PLOT_BUS"]

    if plot_level:
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            logger.debug(
                "Error: Please install \"Matpotlib\" package for plotting function!\n")

        if plot_bus == '':  # no bus specified
            if len(cb_data["Sxfr"][cb_data["pq"]]) != 0:
                idx = argmax(abs(cb_data["Sxfr"][cb_data["pq"]]), axis=0)
                idx = cb_data["pq"][idx]
            else:   # or bus 1 if there are none
                idx = 0
            idx_e = int(cb_data["ppc_target"]["order"]["bus"]["i2e"][idx])

        else:
            idx_e = plot_bus
            if issparse(cb_data["ppc_target"]["order"]["bus"]["e2i"][idx_e]):
                idx = int(cb_data["ppc_target"]["order"]
                          ["bus"]["e2i"][idx_e].toarray())
            else:
                idx = int(cb_data["ppc_target"]["order"]["bus"]["e2i"][idx_e])

            if idx == 0:
                logger.debug(
                    "cpf_default_callback: %d is not a valid bus number for PPOPT[\"CPF_PLOT_BUS\"]\n" % idx_e)

    # -----  FINAL call  -----
    if is_final:
        # assemble results struct
        results = cb_state
        results["max_lam"] = amax(cb_state["lam_c"])
        results["iterations"] = k

        # finish final lambda-V nose curve plot
        if plot_level:
            plt.plot(cb_state["lam_c"],
                     abs(cb_state["V_c"][idx, :]), '-', color=(0.25, 0.25, 1))
            plt.axis([0, amax([1, amax(cb_state["lam_p"]), amax(cb_state["lam_c"])])*1.05, 0,
                      amax([1, amax(abs(cb_state["V_p"][idx])), amax(abs(cb_state["V_c"][idx]))*1.05])])
            plt.pause(sys.float_info.epsilon)
            plt.ioff()

    elif k == 0:
        # initialize state
        cb_state = {
            "V_p": V_p,
            "lam_p": lam_p,
            "V_c": V_c,
            "lam_c": lam_c,
            "iterations": 0
        }

        # initialize lambda-V nose curve plot
        if plot_level:
            plt.ion()
            plt.plot(cb_state["lam_p"],
                     abs(cb_state["V_p"][idx]), '-', color=(0.25, 0.25, 1))
            plt.title('Voltage at Bus %d' % idx_e)
            plt.xlabel('$\\lambda$')
            plt.ylabel('Voltage Magnitude')
            plt.axis([0, amax([1, amax(cb_state["lam_p"]), amax(cb_state["lam_c"])])*1.05, 0,
                      amax([1, amax(abs(cb_state["V_p"][idx])), amax(abs(cb_state["V_c"][idx]))*1.05])])
            plt.pause(sys.float_info.epsilon)

    # -----  ITERATION call  -----
    else:
        # update state
        cb_state["V_p"] = c_[cb_state["V_p"], V_p]
        cb_state["lam_p"] = r_[cb_state["lam_p"], lam_p]
        cb_state["V_c"] = c_[cb_state["V_c"], V_c]
        cb_state["lam_c"] = r_[cb_state["lam_c"], lam_c]
        cb_state["iterations"] = k

        # plot single step of the lambda-V nose curve
        if plot_level > 1:
            plt.plot(r_[cb_state["lam_c"][k-1], cb_state["lam_p"][k]],
                     r_[abs(cb_state["V_c"][idx, k-1]),
                        abs(cb_state["V_p"][idx, k])],
                     '-', color=0.85*array([1, 0.75, 0.75]))
            plt.plot(r_[cb_state["lam_p"][k], cb_state["lam_c"][k]],
                     r_[abs(cb_state["V_p"][idx, k]),
                        abs(cb_state["V_c"][idx, k])],
                     '-', color=0.85*array([0.75, 1, 0.75]))
            plt.plot(cb_state["lam_p"][k], abs(cb_state["V_p"][idx, k]),
                     'x', color=0.85*array([1, 0.75, 0.75]))
            plt.plot(cb_state["lam_c"][k], abs(
                cb_state["V_c"][idx, k]), '-o', markerfacecolor='none', color=(0.25, 0.25, 1))
            plt.axis([0, amax([1, amax(cb_state["lam_p"]), amax(cb_state["lam_c"])])*1.05, 0,
                      amax([1, amax(abs(cb_state["V_p"][idx])), amax(abs(cb_state["V_c"][idx]))*1.05])])
            plt.pause(sys.float_info.epsilon)
            
        if plot_level > 2:
            os.system("pause")

    return cb_state, results



"""
Test system report.
"""
import unittest
import os

import ams


class TestReport(unittest.TestCase):
    """
    Tests for Report class.
    """

    def setUp(self) -> None:
        self.ss = ams.main.load(
            ams.get_case("5bus/pjm5bus_demo.xlsx"),
            default_config=True,
            no_output=True,
        )
        self.expected_report = 'pjm5bus_demo_out.txt'

    def test_no_output(self):
        """
        Test no output.
        """
        self.assertTrue(self.ss.files.no_output)
        self.assertFalse(self.ss.report())

    def test_no_report(self):
        """
        Test report with no solved routine.
        """
        self.ss.files.no_output = False
        self.assertTrue(self.ss.report())
        self.assertTrue(os.path.exists(self.expected_report))

        with open(self.expected_report, "r") as report_file:
            file_contents = report_file.read()
            self.assertNotIn("DCOPF", file_contents)

        os.remove(self.expected_report)

    def test_DCOPF_report(self):
        """
        Test report with DCOPF solved.
        """
        self.ss.files.no_output = False
        self.ss.DCOPF.run(solver='CLARABEL')
        self.assertTrue(self.ss.report())
        self.assertTrue(os.path.exists(self.expected_report))

        with open(self.expected_report, "r") as report_file:
            file_contents = report_file.read()
            self.assertIn("DCOPF", file_contents)

        os.remove(self.expected_report)

    def test_multi_report(self):
        """
        Test report with multiple solved routines.
        """
        self.ss.files.no_output = False
        self.ss.DCOPF.run(solver='CLARABEL')
        self.ss.RTED.run(solver='CLARABEL')
        self.ss.ED.run(solver='CLARABEL')
        self.assertTrue(self.ss.report())
        self.assertTrue(os.path.exists(self.expected_report))

        with open(self.expected_report, "r") as report_file:
            file_contents = report_file.read()
            self.assertIn("DCOPF", file_contents)
            self.assertIn("RTED", file_contents)
            self.assertIn("ED", file_contents)

        os.remove(self.expected_report)



import unittest
import numpy as np

import ams
from ams.shared import skip_unittest_without_MIP


class TestRTED(unittest.TestCase):
    """
    Test methods of `RTED`.
    """

    def setUp(self) -> None:
        self.ss = ams.load(ams.get_case("5bus/pjm5bus_demo.xlsx"),
                           setup=True,
                           default_config=True,
                           no_output=True,
                           )
        self.ss.RTED.run(solver='CLARABEL')

    def test_init(self):
        """
        Test `RTED.init()` method.
        """
        self.ss.RTED.init()
        self.assertTrue(self.ss.RTED.initialized, "RTED initialization failed!")

    def test_trip_gen(self):
        """
        Test generator tripping.
        """
        stg = 'PV_1'
        self.ss.StaticGen.set(src='u', idx=stg, attr='v', value=0)

        self.ss.RTED.update()
        self.ss.RTED.run(solver='CLARABEL')
        self.assertTrue(self.ss.RTED.converged, "RTED did not converge under generator trip!")
        self.assertAlmostEqual(self.ss.RTED.get(src='pg', attr='v', idx=stg),
                               0, places=6,
                               msg="Generator trip does not take effect!")

        self.ss.StaticGen.set(src='u', idx=stg, attr='v', value=1)  # reset

    def test_trip_line(self):
        """
        Test line tripping.
        """
        self.ss.Line.set(src='u', attr='v', idx='Line_3', value=0)

        self.ss.RTED.update()
        self.ss.RTED.run(solver='CLARABEL')
        self.assertTrue(self.ss.RTED.converged, "RTED did not converge under line trip!")
        self.assertAlmostEqual(self.ss.RTED.get(src='plf', attr='v', idx='Line_3'),
                               0, places=6,
                               msg="Line trip does not take effect!")

        self.ss.Line.alter(src='u', idx='Line_3', value=1)

    def test_set_load(self):
        """
        Test setting and tripping load.
        """
        self.ss.RTED.run(solver='CLARABEL')
        obj = self.ss.RTED.obj.v

        # --- set load ---
        self.ss.PQ.set(src='p0', attr='v', idx='PQ_1', value=0.1)
        self.ss.RTED.update()

        self.ss.RTED.run(solver='CLARABEL')
        obj_pqt = self.ss.RTED.obj.v
        self.assertLess(obj_pqt, obj, "Load set does not take effect!")

        # --- trip load ---
        self.ss.PQ.set(src='u', attr='v', idx='PQ_2', value=0)
        self.ss.RTED.update()

        self.ss.RTED.run(solver='CLARABEL')
        obj_pqt2 = self.ss.RTED.obj.v
        self.assertLess(obj_pqt2, obj_pqt, "Load trip does not take effect!")

    def test_dc2ac(self):
        """
        Test `RTED.init()` method.
        """
        self.ss.RTED.dc2ac()
        self.assertTrue(self.ss.RTED.converted, "AC conversion failed!")
        self.assertTrue(self.ss.RTED.exec_time > 0, "Execution time is not greater than 0.")

        stg_idx = self.ss.StaticGen.get_idx()
        pg_rted = self.ss.RTED.get(src='pg', attr='v', idx=stg_idx)
        pg_acopf = self.ss.ACOPF.get(src='pg', attr='v', idx=stg_idx)
        np.testing.assert_almost_equal(pg_rted, pg_acopf, decimal=3)

        bus_idx = self.ss.Bus.get_idx()
        v_rted = self.ss.RTED.get(src='vBus', attr='v', idx=bus_idx)
        v_acopf = self.ss.ACOPF.get(src='vBus', attr='v', idx=bus_idx)
        np.testing.assert_almost_equal(v_rted, v_acopf, decimal=3)

        a_rted = self.ss.RTED.get(src='aBus', attr='v', idx=bus_idx)
        a_acopf = self.ss.ACOPF.get(src='aBus', attr='v', idx=bus_idx)
        np.testing.assert_almost_equal(a_rted, a_acopf, decimal=3)


class TestRTEDES(unittest.TestCase):
    """
    Test routine `RTEDES`.
    """

    def setUp(self) -> None:
        self.ss = ams.load(ams.get_case("5bus/pjm5bus_uced_esd1.xlsx"),
                           setup=True, default_config=True, no_output=True)

    def test_init(self):
        """
        Test initialization.
        """
        self.ss.RTEDES.init()
        self.assertTrue(self.ss.RTEDES.initialized, "RTEDES initialization failed!")

    @skip_unittest_without_MIP
    def test_trip_gen(self):
        """
        Test generator tripping.
        """
        stg = 'PV_1'
        self.ss.StaticGen.set(src='u', idx=stg, attr='v', value=0)

        self.ss.RTEDES.update()
        self.ss.RTEDES.run(solver='SCIP')
        self.assertTrue(self.ss.RTEDES.converged, "RTEDES did not converge under generator trip!")
        self.assertAlmostEqual(self.ss.RTEDES.get(src='pg', attr='v', idx=stg),
                               0, places=6,
                               msg="Generator trip does not take effect!")

        self.ss.StaticGen.set(src='u', idx=stg, attr='v', value=1)

    @skip_unittest_without_MIP
    def test_trip_line(self):
        """
        Test line tripping.
        """
        self.ss.Line.set(src='u', attr='v', idx=3, value=0)

        self.ss.RTEDES.update()
        self.ss.RTEDES.run(solver='SCIP')
        self.assertTrue(self.ss.RTEDES.converged, "RTEDES did not converge under line trip!")
        self.assertAlmostEqual(self.ss.RTEDES.get(src='plf', attr='v', idx=3),
                               0, places=6,
                               msg="Line trip does not take effect!")

        self.ss.Line.alter(src='u', idx=3, value=1)

    @skip_unittest_without_MIP
    def test_set_load(self):
        """
        Test setting and tripping load.
        """
        self.ss.RTEDES.run(solver='SCIP')
        pgs = self.ss.RTEDES.obj.v.sum()

        # --- set load ---
        self.ss.PQ.set(src='p0', attr='v', idx='PQ_1', value=1)
        self.ss.RTEDES.update()

        self.ss.RTEDES.run(solver='SCIP')
        pgs_pqt = self.ss.RTEDES.obj.v.sum()
        self.assertLess(pgs_pqt, pgs, "Load set does not take effect!")

        # --- trip load ---
        self.ss.PQ.set(src='u', attr='v', idx='PQ_2', value=0)
        self.ss.RTEDES.update()

        self.ss.RTEDES.run(solver='SCIP')
        pgs_pqt2 = self.ss.RTEDES.obj.v.sum()
        self.assertLess(pgs_pqt2, pgs_pqt, "Load trip does not take effect!")



"""
Power flow routines.
"""
import logging
from collections import OrderedDict

from ams.pypower import runpf

from ams.io.pypower import system2ppc
from ams.pypower.core import ppoption
from ams.core.param import RParam

from ams.routines.dcpf import DCPF
from ams.opt.omodel import Var

logger = logging.getLogger(__name__)


class PFlow(DCPF):
    """
    AC Power Flow routine.

    Notes
    -----
    1. AC pwoer flow is solved with PYPOWER ``runpf`` function.
    2. AC power flow formulation in AMS style is NOT DONE YET,
       but this does not affect the results
       because the data are passed to PYPOWER for solving.
    """

    def __init__(self, system, config):
        DCPF.__init__(self, system, config)
        self.info = "AC Power Flow"
        self.type = "PF"

        self.config.add(OrderedDict((('qlim', 0),
                                     )))
        self.config.add_extra("_help",
                              qlim="Enforce generator q limits",
                              )
        self.config.add_extra("_alt",
                              qlim=(0, 1, 2),
                              )

        self.qd = RParam(info="reactive power load in system base",
                         name="qd", tex_name=r"q_{d}",
                         unit="p.u.",
                         model="StaticLoad", src="q0",)

        # --- bus ---
        self.vBus = Var(info="bus voltage magnitude",
                        unit="p.u.",
                        name="vBus", tex_name=r"v_{Bus}",
                        model="Bus", src="v",)
        # --- gen ---
        self.qg = Var(info="reactive power generation",
                      unit="p.u.",
                      name="qg", tex_name=r"q_{g}",
                      model="StaticGen", src="q",)
        # NOTE: omit AC power flow formulation here

    def solve(self, method="newton", **kwargs):
        """
        Solve the AC power flow using PYPOWER.
        """
        ppc = system2ppc(self.system)

        method_map = dict(newton=1, fdxb=2, fdbx=3, gauss=4)
        alg = method_map.get(method)
        if alg == 4:
            msg = "Gauss method is not fully tested yet, not recommended!"
            logger.warning(msg)
        if alg is None:
            msg = f"Invalid method `{method}` for PFlow."
            raise ValueError(msg)
        ppopt = ppoption(PF_ALG=alg, ENFORCE_Q_LIMS=self.config.qlim, **kwargs)

        res, sstats = runpf(casedata=ppc, ppopt=ppopt)
        return res, sstats

    def run(self, **kwargs):
        """
        Run AC power flow using PYPOWER.

        Currently, four methods are supported: 'newton', 'fdxb', 'fdbx', 'gauss',
        for Newton's method, fast-decoupled, XB, fast-decoupled, BX, and Gauss-Seidel,
        respectively.

        Note that gauss method is not recommended because it seems to be much
        more slower than the other three methods and not fully tested yet.

        Examples
        --------
        >>> ss = ams.load(ams.get_case('matpower/case14.m'))
        >>> ss.PFlow.run()

        Parameters
        ----------
        force_init : bool
            Force initialization.
        no_code : bool
            Disable showing code.
        method : str
            Method for solving the power flow.

        Returns
        -------
        exit_code : int
            Exit code of the routine.
        """
        return super().run(**kwargs,)



"""
Module for the core of PYPOWER.
"""

from ams.pypower.core.ppoption import ppoption  # NOQA
from ams.pypower.core.pips import pips, pipsver, pipsopf_solver  # NOQA
from ams.pypower.core.ppver import ppver  # NOQA
from ams.pypower.core.solver import ipoptopf_solver  # NOQA



from andes.models.shunt.shunt import ShuntData

from ams.core.model import Model


class Shunt(ShuntData, Model):
    """
    Phasor-domain shunt compensator Model.
    """

    def __init__(self, system=None, config=None):
        ShuntData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'StaticShunt'



"""
ACOPF routines.
"""
import logging
from collections import OrderedDict

from ams.pypower import runopf
from ams.pypower.core import ppoption

from ams.io.pypower import system2ppc
from ams.core.param import RParam

from ams.routines.dcpf import DCPF
from ams.opt.omodel import Var, Constraint, Objective

logger = logging.getLogger(__name__)


class ACOPF(DCPF):
    """
    Standard AC optimal power flow.

    Notes
    -----
    1. ACOPF is solved with PYPOWER ``runopf`` function.
    2. ACOPF formulation in AMS style is NOT DONE YET,
       but this does not affect the results
       because the data are passed to PYPOWER for solving.
    """

    def __init__(self, system, config):
        DCPF.__init__(self, system, config)
        self.info = 'AC Optimal Power Flow'
        self.type = 'ACED'

        self.map1 = OrderedDict()   # ACOPF does not receive
        self.map2 = OrderedDict([
            ('Bus', {
                'vBus': 'v0',
            }),
            ('StaticGen', {
                'pg': 'p0',
            }),
        ])

        # --- params ---
        self.c2 = RParam(info='Gen cost coefficient 2',
                         name='c2', tex_name=r'c_{2}',
                         unit=r'$/(p.u.^2)', model='GCost',
                         indexer='gen', imodel='StaticGen',
                         nonneg=True)
        self.c1 = RParam(info='Gen cost coefficient 1',
                         name='c1', tex_name=r'c_{1}',
                         unit=r'$/(p.u.)', model='GCost',
                         indexer='gen', imodel='StaticGen',)
        self.c0 = RParam(info='Gen cost coefficient 0',
                         name='c0', tex_name=r'c_{0}',
                         unit=r'$', model='GCost',
                         indexer='gen', imodel='StaticGen',
                         no_parse=True)
        self.qd = RParam(info='reactive demand',
                         name='qd', tex_name=r'q_{d}',
                         model='StaticLoad', src='q0',
                         unit='p.u.',)
        # --- bus ---
        self.vBus = Var(info='Bus voltage magnitude',
                        unit='p.u.',
                        name='vBus', tex_name=r'v_{Bus}',
                        src='v', model='Bus',)
        # --- gen ---
        self.qg = Var(info='Gen reactive power',
                      unit='p.u.',
                      name='qg', tex_name=r'q_{g}',
                      model='StaticGen', src='q',)
        # --- constraints ---
        self.pb = Constraint(name='pb',
                             info='power balance',
                             e_str='sum(pd) - sum(pg)',
                             is_eq=True,)
        # TODO: ACOPF formulation
        # --- objective ---
        self.obj = Objective(name='obj',
                             info='total cost',
                             e_str='sum(c2 * pg**2 + c1 * pg + c0)',
                             sense='min',)

    def solve(self, method=None, **kwargs):
        """
        Solve ACOPF using PYPOWER with PIPS.
        """
        ppc = system2ppc(self.system)
        ppopt = ppoption()
        res, sstats = runopf(casedata=ppc, ppopt=ppopt, **kwargs)
        return res, sstats

    def run(self, **kwargs):
        """
        Run ACOPF using PYPOWER with PIPS.
        *args and **kwargs go to `self.solve()`, which are not used yet.

        Examples
        --------
        >>> ss = ams.load(ams.get_case('matpower/case14.m'))
        >>> ss.ACOPF.run()

        Parameters
        ----------
        force_init : bool
            Force initialization.
        no_code : bool
            Disable showing code.
        method : str
            Placeholder for future use.

        Returns
        -------
        exit_code : int
            Exit code of the routine.
        """
        super().run(**kwargs)



"""
PYPOWER module to solve power flow.
"""

import logging

from time import time

import numpy as np

from numpy import flatnonzero as find
from scipy.sparse.linalg import spsolve, splu
from scipy.sparse import hstack, vstack
from scipy.sparse import csr_matrix as c_sparse

from andes.shared import deg2rad, rad2deg

from ams.pypower.core import ppoption
import ams.pypower.utils as putils
from ams.pypower.idx import IDX 
from ams.pypower.io import loadcase
from ams.pypower.eps import EPS
import ams.pypower.routines.opffcns as opfcn

from ams.pypower.make import (makeB, makeBdc, makeSbus, makeYbus, dSbus_dV)

from ams.shared import inf


logger = logging.getLogger(__name__)


def runpf(casedata, ppopt):
    """
    Runs a power flow.

    Runs a power flow (full AC Newton's method by default) and optionally
    returns the solved values in the data matrices, a flag which is True if
    the algorithm was successful in finding a solution, and the elapsed
    time in seconds. All input arguments are optional. If casename is
    provided it specifies the name of the input data file or dict
    containing the power flow data. The default value is 'case9'.

    Parameters
    ----------
    casedata : str, dict, or None, optional
        The name of the input data file or a dict containing the power flow data.
        Default is None.
    ppopt : dict, optional
        PYPOWER options vector. It can be used to specify the solution algorithm
        and output options among other things. Default is None.

    Returns
    -------
    results : dict or None
        Solved power flow results. None if the power flow did not converge.
    sstats : dict
        Solver statistics.

    Notes
    -----
    If the ENFORCE_Q_LIMS option is set to True (default is False), then if any
    generator reactive power limit is violated after running the AC power flow,
    the corresponding bus is converted to a IDX.bus.PQ bus, with Qg at the limit, and
    the case is re-run. The voltage magnitude at the bus will deviate from the
    specified value in order to satisfy the reactive power limit. If the reference
    bus is converted to IDX.bus.PQ, the first remaining IDX.bus.PV bus will be used as the slack
    bus for the next iteration. This may result in the real power output at this
    generator being slightly off from the specified values.

    Enforcing of generator Q limits inspired by contributions from Mu Lin,
    Lincoln University, New Zealand (1/14/05).

    Author
    ------
    Ray Zimmerman (PSERC Cornell)
    """
    sstats = dict(solver_name='PYPOWER',
                  num_iters=1)  # solver stats
    ppopt = ppoption(ppopt)

    # read data
    ppc = loadcase(casedata)

    # add zero columns to branch for flows if needed
    if ppc["branch"].shape[1] < IDX.branch.QT:
        ppc["branch"] = np.c_[ppc["branch"],
                              np.zeros((ppc["branch"].shape[0],
                                        IDX.branch.QT - ppc["branch"].shape[1] + 1))]

    # convert to internal indexing
    ppc = opfcn.ext2int(ppc)
    baseMVA, bus, gen, branch = \
        ppc["baseMVA"], ppc["bus"], ppc["gen"], ppc["branch"]

    # get bus index lists of each type of bus
    ref, pv, pq = putils.bustypes(bus, gen)

    # generator info
    on = find(gen[:, IDX.gen.GEN_STATUS] > 0)  # which generators are on?
    gbus = gen[on, IDX.gen.GEN_BUS].astype(int)  # what buses are they at?

    # -----  run the power flow  -----
    t0 = time()

    if ppopt["PF_DC"]:                               # DC formulation
        # initial state
        Va0 = bus[:, IDX.bus.VA] * deg2rad

        # build B matrices and phase shift injections
        B, Bf, Pbusinj, Pfinj, _ = makeBdc(baseMVA, bus, branch)

        # compute complex bus power injections [generation - load]
        # adjusted for phase shifters and real shunts
        Pbus = makeSbus(baseMVA, bus, gen).real - Pbusinj - bus[:, IDX.bus.GS] / baseMVA

        # "run" the power flow
        Va = dcpf(B, Pbus, Va0, ref, pv, pq)

        # update data matrices with solution
        branch[:, [IDX.branch.QF, IDX.branch.QT]] = np.zeros((branch.shape[0], 2))
        branch[:, IDX.branch.PF] = (Bf * Va + Pfinj) * baseMVA
        branch[:, IDX.branch.PT] = -branch[:, IDX.branch.PF]
        bus[:, IDX.bus.VM] = np.ones(bus.shape[0])
        bus[:, IDX.bus.VA] = Va * rad2deg
        # update Pg for slack generator (1st gen at ref bus)
        # (note: other gens at ref bus are accounted for in Pbus)
        # Pg = Pinj + Pload + Gs
        # newPg = oldPg + newPinj - oldPinj
        refgen = np.zeros(len(ref), dtype=int)
        for k in range(len(ref)):
            temp = find(gbus == ref[k])
            refgen[k] = on[temp[0]]
        gen[refgen, IDX.gen.PG] = gen[refgen, IDX.gen.PG] + (B[ref, :] * Va - Pbus[ref]) * baseMVA

        success = 1
    else:  # AC formulation
        method_map = {1: 'Newton', 2: 'fast-decoupled, XB',
                      3: 'fast-decoupled, BX', 4: 'Gauss-Seidel'}
        alg = method_map.get(ppopt['PF_ALG'])
        sstats['solver_name'] = f'PYPOWER-{alg}'
        logger.debug(f"Solution method: {alg}'s method.")
        if alg is None:
            logger.debug('Only Newton\'s method, fast-decoupled, and '
                         'Gauss-Seidel power flow algorithms currently '
                         'implemented.\n')
            raise ValueError

        # initial state
        # V0    = np.ones(bus.shape[0])            ## flat start
        V0 = bus[:, IDX.bus.VM] * np.exp(1j * deg2rad * bus[:, IDX.bus.VA])
        vcb = np.ones(V0.shape)    # create mask of voltage-controlled buses
        vcb[pq] = 0     # exclude IDX.bus.PQ buses
        k = find(vcb[gbus])     # in-service gens at v-c buses
        V0[gbus[k]] = gen[on[k], IDX.gen.VG] / abs(V0[gbus[k]]) * V0[gbus[k]]

        if ppopt["ENFORCE_Q_LIMS"]:
            ref0 = ref  # save index and angle of
            Varef0 = bus[ref0, IDX.bus.VA]  # original reference bus(es)
            limited = []  # list of indices of gens @ Q lims
            fixedQg = np.zeros(gen.shape[0])  # Qg of gens at Q limits

        # build admittance matrices
        Ybus, Yf, Yt = makeYbus(baseMVA, bus, branch)

        repeat = True
        while repeat:
            # compute complex bus power injections [generation - load]
            Sbus = makeSbus(baseMVA, bus, gen)

            # run the power flow
            if ppopt['PF_ALG'] == 1:
                V, success, sstats['num_iters'] = newtonpf(Ybus, Sbus, V0, ref, pv, pq, ppopt)
            elif ppopt['PF_ALG'] in (2, 3):
                Bp, Bpp = makeB(baseMVA, bus, branch, alg)
                V, success, sstats['num_iters'] = fdpf(Ybus, Sbus, V0, Bp, Bpp, ref, pv, pq, ppopt)
            elif ppopt['PF_ALG'] == 4:
                V, success, sstats['num_iters'] = gausspf(Ybus, Sbus, V0, ref, pv, pq, ppopt)
            else:
                pass
                # update data matrices with solution
            bus, gen, branch = pfsoln(baseMVA, bus, gen, branch, Ybus, Yf, Yt, V, ref, pv, pq)

            enforce_q_lims = ppopt["ENFORCE_Q_LIMS"]
            if ppopt["ENFORCE_Q_LIMS"]:  # enforce generator Q limits
                # find gens with violated Q constraints
                gen_status = gen[:, IDX.gen.GEN_STATUS] > 0
                qg_max_lim = gen[:, IDX.gen.QG] > gen[:, IDX.gen.QMAX] + ppopt["OPF_VIOLATION"]
                qg_min_lim = gen[:, IDX.gen.QG] < gen[:, IDX.gen.QMIN] - ppopt["OPF_VIOLATION"]

                mx = find(gen_status & qg_max_lim)
                mn = find(gen_status & qg_min_lim)

                if len(mx) > 0 or len(mn) > 0:  # we have some Q limit violations
                    # first check for INFEASIBILITY (all remaining gens violating)
                    gen_violate = np.union1d(mx, mn)
                    gen_bus = gen[:, IDX.gen.GEN_BUS].astype(int)
                    cond_pv = bus[gen_bus, IDX.bus.BUS_TYPE] == IDX.bus.PV
                    cond_ref = bus[gen_bus, IDX.bus.BUS_TYPE] == IDX.bus.REF
                    gen_remain = find(gen_status & (cond_pv | cond_ref))
                    msg = f'gen_violate = {gen_violate}\nremaining = {gen_remain}'
                    logger.debug(msg)
                    # TODO: condition can be improved, but not sure now
                    if len(gen_remain) <= 0:
                        msg = f'Infeasible: remaining {len(gen_violate)} gens exceed Q limits.'
                        logger.warning(msg)
                        success = 0
                        break

                    # one at a time?
                    if ppopt["ENFORCE_Q_LIMS"] == 2:  # fix largest violation, ignore the rest
                        k = np.argmax(np.r_[gen[mx, IDX.gen.QG] - gen[mx, IDX.gen.QMAX],
                                      gen[mn, IDX.gen.QMIN] - gen[mn, IDX.gen.QG]])
                        if k > len(mx):
                            mn = mn[k - len(mx)]
                            mx = []
                        else:
                            mx = mx[k]
                            mn = []

                    if len(mx) > 0:
                        msg = 'Following Gen convert to PQ bus because'
                        msg += ' exceed Q upper limits:\n'
                        msg += ', '.join(str(i + 1) for i in mx)
                        logger.debug(msg)
                    if len(mn) > 0:
                        msg = 'Following Gen convert to PQ bus because'
                        msg += ' exceed Q lower limits:\n'
                        msg += ', '.join(str(i + 1) for i in mn)
                        logger.debug(msg)

                    # save corresponding limit values
                    fixedQg[mx] = gen[mx, IDX.gen.QMAX]
                    fixedQg[mn] = gen[mn, IDX.gen.QMIN]
                    mx = np.r_[mx, mn].astype(int)

                    # convert to PQ bus
                    # Convert generators PV bus to PQ bus
                    for i in range(len(mx)):
                        idx = mx[i]
                        gen[idx, IDX.gen.QG] = fixedQg[idx]  # Set Qg to binding
                        gen[idx, IDX.gen.GEN_STATUS] = 0  # Temporarily turn off generator
                        bi = int(gen[idx, IDX.gen.GEN_BUS])  # Get the bus index
                        bus[bi, [IDX.bus.PD, IDX.bus.QD]] -= gen[idx, [IDX.gen.PG, IDX.gen.QG]]  # Adjust load

                    if len(ref) > 1 and any(bus[gen[mx, IDX.gen.GEN_BUS], IDX.bus.BUS_TYPE] == IDX.bus.REF):
                        raise ValueError("PYPOWER cannot enforce Q limits for systems with multiple slack buses. "
                                         "Please ensure there is only one slack bus in the system.")

                    # set bus type to PQ
                    bus[gen[mx, IDX.gen.GEN_BUS].astype(int), IDX.bus.BUS_TYPE] = IDX.bus.PQ

                    # update bus index lists of each type of bus
                    ref_temp = ref
                    ref, pv, pq = putils.bustypes(bus, gen)

                    # previous line can modify lists to select new REF bus
                    # if there was none, so we should update bus with these
                    # just to keep them consistent
                    if ref != ref_temp:
                        bus[ref, IDX.bus.BUS_TYPE] = IDX.bus.REF
                        bus[pv, IDX.bus.BUS_TYPE] = pv
                        logger.debug(f'Bus<{ref[0]}> is new slack bus')

                    limited = np.r_[limited, mx].astype(int)
                else:
                    repeat = 0  # no more generator Q limits violated
            else:
                repeat = 0  # don't enforce generator Q limits, once is enough

            if ppopt["ENFORCE_Q_LIMS"] and len(limited) > 0:
                # Restore injections from limited gens [those at Q limits]
                for i in range(len(limited)):
                    idx = limited[i]
                    gen[idx, IDX.gen.QG] = fixedQg[idx]  # Restore Qg value
                    bi = int(gen[idx, IDX.gen.GEN_BUS])  # Get the bus index
                    bus[bi, [IDX.bus.PD, IDX.bus.QD]] += gen[idx,
                                                             [IDX.gen.PG, IDX.gen.QG]]  # Re-adjust load
                    gen[idx, IDX.gen.GEN_STATUS] = 1  # Turn generator back on

                if ref != ref0:
                    # adjust voltage angles to make original ref bus correct
                    bus[:, IDX.bus.VA] = bus[:, IDX.bus.VA] - bus[ref0, IDX.bus.VA] + Varef0

    ppc["et"] = time() - t0
    ppc["success"] = success

    # -----  output results  -----
    # convert back to original bus numbering & print results
    ppc["bus"], ppc["gen"], ppc["branch"] = bus, gen, branch
    results = opfcn.int2ext(ppc)

    # zero out result fields of out-of-service gens & branches
    if len(results["order"]["gen"]["status"]["off"]) > 0:
        results["gen"][np.ix_(results["order"]["gen"]["status"]["off"], [IDX.gen.PG, IDX.gen.QG])] = 0

    if len(results["order"]["branch"]["status"]["off"]) > 0:
        results["branch"][
            np.ix_(
                results["order"]["branch"]["status"]["off"],
                [IDX.branch.PF,
                 IDX.branch.QF,
                 IDX.branch.PT,
                 IDX.branch.QT])] = 0

    return results, sstats


def dcpf(B, Pbus, Va0, ref, pv, pq):
    """
    Solves a DC power flow.

    Solves for the bus voltage angles at all but the reference bus, given the
    full system B matrix and the vector of bus real power injections, the
    initial vector of bus voltage angles (rad), and column vectors with
    the lists of bus indices for the swing bus, PV buses, and PQ buses,
    respectively.

    Parameters
    ----------
    B : ndarray
        The system B matrix.
    Pbus : ndarray
        Vector of bus real power injections.
    Va0 : ndarray
        Initial vector of bus voltage angles (in radians).
    ref : int
        Index of the reference bus.
    pv : ndarray
        List of bus indices for IDX.bus.PV buses.
    pq : ndarray
        List of bus indices for IDX.bus.PQ buses.

    Returns
    -------
    Va : ndarray
        Vector of bus voltage angles in radians.

    Author
    ------
    Carlos E. Murillo-Sanchez (PSERC Cornell & Universidad Autonoma de Manizales)

    Ray Zimmerman (PSERC Cornell)
    """
    pvpq = np.r_[pv, pq][np.newaxis, :]

    # initialize result vector
    Va = np.copy(Va0)

    # update angles for non-reference buses
    Va[pvpq] = spsolve(B[pvpq.T, pvpq], np.transpose(Pbus[pvpq] - B[pvpq.T, ref] * Va0[ref]))

    return Va


def newtonpf(Ybus, Sbus, V0, ref, pv, pq, ppopt):
    """
    Solves the power flow using a full Newton's method.

    Parameters
    ----------
    Ybus : array-like
        Full system admittance matrix (for all buses).
    Sbus : array-like
        Complex bus power injection vector (for all buses).
    V0 : array-like
        Initial vector of complex bus voltages.
    ref : int
        Index of the swing bus.
    pv : list of int
        List of bus indices for PV buses.
    pq : list of int
        List of bus indices for PQ buses.
    ppopt : dict, optional
        PYPOWER options vector which can be used to set the termination tolerance,
        maximum number of iterations, and output options (see ppoption for details).
        Default is None, which uses default options.

    Returns
    -------
    V : array-like
        Final complex voltages.
    converged : bool
        Flag indicating whether the power flow converged or not.
    i : int
        Number of iterations performed.

    Notes
    -----
    Solves for bus voltages given the full system admittance matrix (for
    all buses), the complex bus power injection vector (for all buses),
    the initial vector of complex bus voltages, and column vectors with
    the lists of bus indices for the swing bus, PV buses, and PQ buses,
    respectively. The bus voltage vector contains the set point for
    generator (including the reference bus) buses, and the reference angle
    of the swing bus, as well as an initial guess for remaining magnitudes
    and angles. Uses default options if the ppopt parameter is not given.

    Author
    ------
    Ray Zimmerman (PSERC Cornell)
    """
    ppopt = ppoption(ppopt)
    # options
    tol = ppopt['PF_TOL']
    max_it = ppopt['PF_MAX_IT']
    verbose = ppopt['VERBOSE']

    # initialize
    converged = 0
    i = 0
    V = V0
    Va = np.angle(V)
    Vm = abs(V)

    # set up indexing for updating V
    pvpq = np.r_[pv, pq]
    npv = len(pv)
    npq = len(pq)
    j1 = 0
    j2 = npv  # j1:j2 - V angle of pv buses
    j3 = j2
    j4 = j2 + npq  # j3:j4 - V angle of pq buses
    j5 = j4
    j6 = j4 + npq  # j5:j6 - V mag of pq buses

    # evaluate F(x0)
    mis = V * np.conj(Ybus * V) - Sbus
    F = np.r_[mis[pv].real,
              mis[pq].real,
              mis[pq].imag]

    # check tolerance
    normF = np.linalg.norm(F, inf)
    logger.info('%2d: |F(x)| = %.10g', i, normF)
    converged = normF < tol

    # do Newton iterations
    while (not converged and i < max_it):
        # update iteration counter
        i = i + 1

        # evaluate Jacobian
        dS_dVm, dS_dVa = dSbus_dV(Ybus, V)

        J11 = dS_dVa[np.array([pvpq]).T, pvpq].real
        J12 = dS_dVm[np.array([pvpq]).T, pq].real
        J21 = dS_dVa[np.array([pq]).T, pvpq].imag
        J22 = dS_dVm[np.array([pq]).T, pq].imag

        J = vstack([
            hstack([J11, J12]),
            hstack([J21, J22])
        ], format="csr")

        # compute update step
        dx = -1 * spsolve(J, F)

        # update voltage
        if npv:
            Va[pv] = Va[pv] + dx[j1:j2]
        if npq:
            Va[pq] = Va[pq] + dx[j3:j4]
            Vm[pq] = Vm[pq] + dx[j5:j6]
        V = Vm * np.exp(1j * Va)
        Vm = abs(V)  # update Vm and Va again in case
        Va = np.angle(V)  # we wrapped around with a negative Vm

        # evalute F(x)
        mis = V * np.conj(Ybus * V) - Sbus
        F = np.r_[mis[pv].real,
                  mis[pq].real,
                  mis[pq].imag]

        # check for convergence
        normF = np.linalg.norm(F, inf)
        logger.info('%2d: |F(x)| = %.10g', i, normF)
        converged = normF < tol

    return V, converged, i


def pfsoln(baseMVA, bus0, gen0, branch0, Ybus, Yf, Yt, V, ref, pv, pq):
    """
    Updates bus, gen, and branch data structures to match power flow solution.

    This function takes the following inputs and updates the input data structures:

    Parameters
    ----------
    baseMVA : float
        Base power in MVA.
    bus0 : ndarray
        Initial bus data structure.
    gen0 : ndarray
        Initial gen data structure.
    branch0 : ndarray
        Initial branch data structure.
    Ybus : sparse matrix
        Admittance matrix.
    Yf : sparse matrix
        Admittance matrix for "from" end of branches.
    Yt : sparse matrix
        Admittance matrix for "to" end of branches.
    V : ndarray
        Bus voltage magnitude in per unit.
    ref : int
        Reference bus index.
    pv : ndarray
        PV bus indices.
    pq : ndarray
        PQ bus indices.

    Returns
    -------
    bus : ndarray
        Updated bus data structure.
    gen : ndarray
        Updated gen data structure.
    branch : ndarray
        Updated branch data structure.

    Author
    ------
    Ray Zimmerman (PSERC Cornell)
    """
    # initialize return values
    bus = bus0
    gen = gen0
    branch = branch0

    # ----- update bus voltages -----
    bus[:, IDX.bus.VM] = abs(V)
    bus[:, IDX.bus.VA] = np.angle(V) * rad2deg

    # ----- update Qg for all gens and Pg for slack bus(es) -----
    # generator info
    on = find(gen[:, IDX.gen.GEN_STATUS] > 0)  # which generators are on?
    gbus = gen[on, IDX.gen.GEN_BUS].astype(int)  # what buses are they at?

    # compute total injected bus powers
    Sbus = V[gbus] * np.conj(Ybus[gbus, :] * V)

    # update Qg for all generators
    gen[:, IDX.gen.QG] = np.zeros(gen.shape[0])  # zero out all Qg
    gen[on, IDX.gen.QG] = Sbus.imag * baseMVA + bus[gbus, IDX.bus.QD]  # inj Q + local Qd
    # ... at this point any buses with more than one generator will have
    # the total Q dispatch for the bus assigned to each generator. This
    # must be split between them. We do it first equally, then in proportion
    # to the reactive range of the generator.

    if len(on) > 1:
        # build connection matrix, element i, j is 1 if gen on(i) at bus j is ON
        nb = bus.shape[0]
        ngon = on.shape[0]
        Cg = c_sparse((np.ones(ngon), (range(ngon), gbus)), (ngon, nb))

        # divide Qg by number of generators at the bus to distribute equally
        ngg = Cg * Cg.sum(0).T  # ngon x 1, number of gens at this gen's bus
        ngg = np.asarray(ngg).flatten()  # 1D array
        gen[on, IDX.gen.QG] = gen[on, IDX.gen.QG] / ngg

        # divide proportionally
        Cmin = c_sparse((gen[on, IDX.gen.QMIN], (range(ngon), gbus)), (ngon, nb))
        Cmax = c_sparse((gen[on, IDX.gen.QMAX], (range(ngon), gbus)), (ngon, nb))
        Qg_tot = Cg.T * gen[on, IDX.gen.QG]  # nb x 1 vector of total Qg at each bus
        Qg_min = Cmin.sum(0).T  # nb x 1 vector of min total Qg at each bus
        Qg_max = Cmax.sum(0).T  # nb x 1 vector of max total Qg at each bus
        Qg_min = np.asarray(Qg_min).flatten()  # 1D array
        Qg_max = np.asarray(Qg_max).flatten()  # 1D array
        # gens at buses with Qg range = 0
        ig = find(Cg * Qg_min == Cg * Qg_max)
        Qg_save = gen[on[ig], IDX.gen.QG]
        gen[on, IDX.gen.QG] = gen[on, IDX.gen.QMIN] + \
            (Cg * ((Qg_tot - Qg_min) / (Qg_max - Qg_min + EPS))) * \
            (gen[on, IDX.gen.QMAX] - gen[on, IDX.gen.QMIN])  # ^ avoid div by 0
        gen[on[ig], IDX.gen.QG] = Qg_save  # (terms are mult by 0 anyway)

    # update Pg for slack bus(es)
    # inj P + local Pd
    for k in range(len(ref)):
        refgen = find(gbus == ref[k])  # which is(are) the reference gen(s)?
        gen[on[refgen[0]], IDX.gen.PG] = \
            Sbus[refgen[0]].real * baseMVA + bus[ref[k], IDX.bus.PD]
        if len(refgen) > 1:  # more than one generator at this ref bus
            # subtract off what is generated by other gens at this bus
            gen[on[refgen[0]], IDX.gen.PG] = \
                gen[on[refgen[0]], IDX.gen.PG] - sum(gen[on[refgen[1:len(refgen)]], IDX.gen.PG])

    # ----- update/compute branch power flows -----
    out = find(branch[:, IDX.branch.BR_STATUS] == 0)  # out-of-service branches
    br = find(branch[:, IDX.branch.BR_STATUS]).astype(int)  # in-service branches

    # complex power at "from" bus
    Sf = V[branch[br, IDX.branch.F_BUS].astype(int)] * np.conj(Yf[br, :] * V) * baseMVA
    # complex power injected at "to" bus
    St = V[branch[br, IDX.branch.T_BUS].astype(int)] * np.conj(Yt[br, :] * V) * baseMVA
    branch[np.ix_(br, [IDX.branch.PF, IDX.branch.QF, IDX.branch.PT, IDX.branch.QT])
           ] = np.c_[Sf.real, Sf.imag, St.real, St.imag]
    branch[np.ix_(out, [IDX.branch.PF, IDX.branch.QF, IDX.branch.PT, IDX.branch.QT])
           ] = np.zeros((len(out), 4))

    return bus, gen, branch


def gausspf(Ybus, Sbus, V0, ref, pv, pq, ppopt):
    """
    Solves the power flow using a Gauss-Seidel method.
    This method seems to be much more slower than Newton's method,
    not fully checked yet.

    Parameters
    ----------
    Ybus : array-like
        Full system admittance matrix (for all buses).
    Sbus : array-like
        Complex bus power injection vector (for all buses).
    V0 : array-like
        Initial vector of complex bus voltages.
    ref : int
        Index of the swing bus.
    pv : list of int
        List of bus indices for PV buses.
    pq : list of int
        List of bus indices for PQ buses.
    ppopt : dict
        PYPOWER options vector which can be used to set the termination tolerance,
        maximum number of iterations, and output options (see ppoption for details).

    Returns
    -------
    V : array-like
        Final complex voltages.
    converged : bool
        Flag indicating whether the power flow converged or not.
    i : int
        Number of iterations performed.

    Notes
    -----
    Solves for bus voltages given the full system admittance matrix (for
    all buses), the complex bus power injection vector (for all buses),
    the initial vector of complex bus voltages, and column vectors with
    the lists of bus indices for the swing bus, PV buses, and PQ buses,
    respectively. The bus voltage vector contains the set point for
    generator (including the reference bus) buses, and the reference angle
    of the swing bus, as well as an initial guess for remaining magnitudes
    and angles. Uses default options if the ppopt parameter is not given.

    Author
    ------
    Ray Zimmerman (PSERC Cornell)

    Alberto Borghetti (University of Bologna, Italy)
    """
    ppopt = ppoption(ppopt)
    # options
    tol = ppopt['PF_TOL']
    max_it = ppopt['PF_MAX_IT_GS']

    # initialize
    converged = 0
    i = 0
    V = V0.copy()
    # Va = angle(V)
    Vm = abs(V)

    # set up indexing for updating V
    npv = len(pv)
    npq = len(pq)
    pvpq = np.r_[pv, pq]

    # evaluate F(x0)
    mis = V * np.conj(Ybus * V) - Sbus
    F = np.r_[mis[pvpq].real,
              mis[pq].imag]

    # check tolerance
    normF = np.linalg.norm(F, inf)
    logger.info('%d: |F(x)| = %.10g', i, normF)
    converged = normF < tol

    # do Gauss-Seidel iterations
    while (not converged and i < max_it):
        # update iteration counter
        i = i + 1

        # update voltage
        # at PQ buses
        for k in pq[list(range(npq))]:
            tmp = (np.conj(Sbus[k] / V[k]) - Ybus[k, :] * V) / Ybus[k, k]
            V[k] = V[k] + tmp.item()

        # at PV buses
        if npv:
            for k in pv[list(range(npv))]:
                tmp = (V[k] * np.conj(Ybus[k, :] * V)).imag
                Sbus[k] = Sbus[k].real + 1j * tmp.item()
                tmp = (np.conj(Sbus[k] / V[k]) - Ybus[k, :] * V) / Ybus[k, k]
                V[k] = V[k] + tmp.item()
#               V[k] = Vm[k] * V[k] / abs(V[k])
            V[pv] = Vm[pv] * V[pv] / abs(V[pv])

        # evalute F(x)
        mis = V * np.conj(Ybus * V) - Sbus
        F = np.r_[mis[pv].real,
                  mis[pq].real,
                  mis[pq].imag]

        # check for convergence
        normF = np.linalg.norm(F, inf)
        logger.info('%d: |F(x)| = %.10g', i, normF)
        converged = normF < tol

    return V, converged, i


def fdpf(Ybus, Sbus, V0, Bp, Bpp, ref, pv, pq, ppopt):
    """
    Solves the power flow using a fast decoupled method.

    Solves for bus voltages given the full system admittance matrix (for
    all buses), the complex bus power injection vector (for all buses),
    the initial vector of complex bus voltages, the FDPF matrices B prime
    and B double prime, and column vectors with the lists of bus indices
    for the swing bus, PV buses, and PQ buses, respectively. The bus voltage
    vector contains the set point for generator (including the reference bus)
    buses, and the reference angle of the swing bus, as well as an initial
    guess for remaining magnitudes and angles. `ppopt` is a PYPOWER options
    vector which can be used to set the termination tolerance, maximum
    number of iterations, and output options (see `ppoption` for details).
    Uses default options if this parameter is not given. Returns the
    final complex voltages, a flag which indicates whether it converged
    or not, and the number of iterations performed.

    Parameters
    ----------
    Ybus : ndarray
        Full system admittance matrix (for all buses).
    Sbus : ndarray
        Complex bus power injection vector (for all buses).
    V0 : ndarray
        Initial vector of complex bus voltages.
    Bp : ndarray
        FDPF matrix B prime.
    Bpp : ndarray
        FDPF matrix B double prime.
    ref : int
        Index of the reference bus.
    pv : ndarray
        List of bus indices for PV buses.
    pq : ndarray
        List of bus indices for PQ buses.
    ppopt : dict
        PYPOWER options vector.

    Returns
    -------
    V : ndarray
        Final complex voltages.
    converged : bool
        Flag indicating whether the power flow converged.
    i : int
        Number of iterations performed.

    Author
    ------
    Ray Zimmerman (PSERC Cornell)
    """
    ppopt = ppoption(ppopt)
    # options
    tol = ppopt['PF_TOL']
    max_it = ppopt['PF_MAX_IT_FD']
    verbose = ppopt['VERBOSE']

    # initialize
    converged = 0
    i = 0
    V = V0
    Va = np.angle(V)
    Vm = abs(V)

    # set up indexing for updating V
    # npv = len(pv)
    # npq = len(pq)
    pvpq = np.r_[pv, pq]

    # evaluate initial mismatch
    mis = (V * np.conj(Ybus * V) - Sbus) / Vm
    P = mis[pvpq].real
    Q = mis[pq].imag

    # check tolerance
    normP = np.linalg.norm(P, inf)
    normQ = np.linalg.norm(Q, inf)
    logger.info(f'{i} --: |P| = {normP:.10g}, |Q| = {normQ:.10g}')
    converged = normP < tol and normQ < tol

    # reduce B matrices
    Bp = Bp[np.array([pvpq]).T, pvpq].tocsc()  # splu requires a CSC matrix
    Bpp = Bpp[np.array([pq]).T, pq].tocsc()

    # factor B matrices
    Bp_solver = splu(Bp)
    Bpp_solver = splu(Bpp)

    # do P and Q iterations
    while (not converged and i < max_it):
        # update iteration counter
        i = i + 1

        # -----  do P iteration, update Va  -----
        dVa = -Bp_solver.solve(P)

        # update voltage
        Va[pvpq] = Va[pvpq] + dVa
        V = Vm * np.exp(1j * Va)

        # evalute mismatch
        mis = (V * np.conj(Ybus * V) - Sbus) / Vm
        P = mis[pvpq].real
        Q = mis[pq].imag

        # check tolerance
        normP = np.linalg.norm(P, inf)
        normQ = np.linalg.norm(Q, inf)
        logger.info(f'{i} P: |P| = {normP:.10g}, |Q| = {normQ:.10g}')
        if normP < tol and normQ < tol:
            converged = 1
            break

        # -----  do Q iteration, update Vm  -----
        dVm = -Bpp_solver.solve(Q)

        # update voltage
        Vm[pq] = Vm[pq] + dVm
        V = Vm * np.exp(1j * Va)

        # evalute mismatch
        mis = (V * np.conj(Ybus * V) - Sbus) / Vm
        P = mis[pvpq].real
        Q = mis[pq].imag

        # check tolerance
        normP = np.linalg.norm(P, inf)
        normQ = np.linalg.norm(Q, inf)
        logger.info(f'{i} Q: |P| = {normP:.10g}, |Q| = {normQ:.10g}')
        converged = normP < tol and normQ < tol
        if converged:
            break

    return V, converged, i



from andes.utils.misc import elapsed

from ams.utils import paths  # NOQA
from ams.utils.paths import get_case  # NOQA


def timer(func):
    def wrapper(*args, **kwargs):
        t0, _ = elapsed()
        result = func(*args, **kwargs)
        _, s = elapsed(t0)
        return result, s
    return wrapper


def create_entry(*fields, three_params=True):
    """
    Helper function to create a list of fields for a model entry.

    Parameters
    ----------
    fields : tuple
        Additional fields to include in the list.
    three_params : bool, optional
        Whether to include 'idx', 'u', 'name' in the list (default is True).

    Returns
    -------
    list
        A list of fields for the model entry.
    """
    base_fields = ['idx', 'u', 'name'] if three_params else []
    return base_fields + list(fields)


def pretty_long_message(message, prefix="", max_length=80):
    """
    Pretty print a long message.

    Parameters
    ----------
    message : str
        The message to format.
    prefix : str, optional
        A prefix to add to each line of the message.
    max_length : int, optional
        The maximum length of each line.

    Returns
    -------
    str
        The formatted message.
    """
    if len(message) <= max_length:
        return message
    else:
        lines = [message[i:i+max_length] for i in range(0, len(message), max_length)]
        formatted_message = lines[0] + "\n" + "\n".join([prefix + line for line in lines[1:]])
        return formatted_message



import unittest
import numpy as np

import ams

ams.config_logger(stream_level=40)


class TestAddressing(unittest.TestCase):
    """
    Tests for addressing.
    """

    def test_ieee14_address(self):
        """
        Test IEEE14 address.
        """

        # FIXME: why there will be case parsing information using ams.system.example()?
        ss = ams.system.example()

        # bus variable indices (internal)
        np.testing.assert_array_equal(ss.Bus.a.a,
                                      np.arange(0, ss.Bus.n, 1))
        np.testing.assert_array_equal(ss.Bus.v.a,
                                      np.arange(ss.Bus.n, 2*ss.Bus.n, 1))

        # external variable indices
        np.testing.assert_array_equal(ss.PV.ud.a,
                                      np.array([31, 32, 33, 34]))
        np.testing.assert_array_equal(ss.PV.p.a,
                                      np.array([35, 36, 37, 38]))
        np.testing.assert_array_equal(ss.PV.q.a,
                                      np.array([39, 40, 41, 42]))
        np.testing.assert_array_equal(ss.Slack.ud.a,
                                      np.array([28]))
        np.testing.assert_array_equal(ss.Slack.p.a,
                                      np.array([29]))
        np.testing.assert_array_equal(ss.Slack.q.a,
                                      np.array([30]))



"""
Module for interfacing ANDES
"""

import os
import logging
from collections import OrderedDict, Counter

from andes.utils.misc import elapsed
from andes.system import System as andes_System

from ams.utils import create_entry
from ams.io import input_formats
from ams.shared import nan, pd, np

logger = logging.getLogger(__name__)


# Models used in ANDES PFlow
pflow_dict = OrderedDict([
    ('Bus', create_entry('Vn', 'vmax', 'vmin', 'v0', 'a0',
                         'xcoord', 'ycoord', 'area', 'zone',
                         'owner')),
    ('PQ', create_entry('bus', 'Vn', 'p0', 'q0', 'vmax',
                        'vmin', 'owner')),
    ('PV', create_entry('Sn', 'Vn', 'bus', 'busr', 'p0', 'q0',
                        'pmax', 'pmin', 'qmax', 'qmin',
                        'v0', 'vmax', 'vmin', 'ra', 'xs')),
    ('Slack', create_entry('Sn', 'Vn', 'bus', 'busr', 'p0', 'q0',
                           'pmax', 'pmin', 'qmax', 'qmin',
                           'v0', 'vmax', 'vmin', 'ra', 'xs', 'a0')),
    ('Shunt', create_entry('Sn', 'Vn', 'bus', 'g', 'b', 'fn')),
    ('Line', create_entry('bus1', 'bus2', 'Sn', 'fn', 'Vn1', 'Vn2',
                          'r', 'x', 'b', 'g', 'b1', 'g1', 'b2', 'g2',
                          'trans', 'tap', 'phi', 'rate_a', 'rate_b',
                          'rate_c', 'owner', 'xcoord', 'ycoord')),
    ('Area', create_entry()),
    ('Jumper', create_entry('bus1', 'bus2')),
])

# dict for guessing dynamic models given its idx
idx_guess = {'rego': 'RenGovernor',
             'ree': 'RenExciter',
             'rea': 'RenAerodynamics',
             'rep': 'RenPitch',
             'busf': 'BusFreq',
             'zone': 'Region',
             'gen': 'StaticGen',
             'pq': 'PQ', }


def to_andes(system, addfile=None,
             setup=False, no_output=False,
             default_config=True,
             verify=True, tol=1e-3):
    """
    Convert the AMS system to an ANDES system.

    A preferred dynamic system file to be added has following features:
    1. The file contains both power flow and dynamic models.
    2. The file can run in ANDES natively.
    3. Power flow models are in the same shape as the AMS system.
    4. Dynamic models, if any, are in the same shape as the AMS system.

    This function is wrapped as the ``System`` class method ``to_andes()``.
    Using the file conversion ``to_andes()`` will automatically
    link the AMS system instance to the converted ANDES system instance
    in the AMS system attribute ``dyn``.

    It should be noted that detailed dynamic simualtion requires extra
    dynamic models to be added to the ANDES system, which can be passed
    through the ``addfile`` argument.

    Parameters
    ----------
    system : System
        The AMS system to be converted to ANDES format.
    addfile : str, optional
        The additional file to be converted to ANDES dynamic mdoels.
    setup : bool, optional
        Whether to call `setup()` after the conversion. Default is True.
    no_output : bool, optional
        To ANDES system.
    default_config : bool, optional
        To ANDES system.
    verify : bool
        If True, the converted ANDES system will be verified with the source
        AMS system using AC power flow.
    tol : float
        The tolerance of error.

    Returns
    -------
    adsys : andes.system.System
        The converted ANDES system.

    Examples
    --------
    >>> import ams
    >>> import andes
    >>> sp = ams.load(ams.get_case('ieee14/ieee14_uced.xlsx'), setup=True)
    >>> sa = sp.to_andes(addfile=andes.get_case('ieee14/ieee14_full.xlsx'),
    ...                  setup=False, overwrite=True, no_output=True)

    Notes
    -----
    1. Power flow models in the addfile will be skipped and only dynamic models will be used.
    2. The addfile format is guessed based on the file extension. Currently only ``xlsx`` is supported.
    3. Index in the addfile is automatically adjusted when necessary.
    """
    t0, _ = elapsed()

    adsys = andes_System(no_output=no_output,
                         default_config=default_config)
    # FIXME: is there a systematic way to do this? Other config might be needed
    adsys.config.freq = system.config.freq
    adsys.config.mva = system.config.mva

    for mdl_name, mdl_cols in pflow_dict.items():
        mdl = getattr(system, mdl_name)
        mdl.cache.refresh("df_in")  # refresh cache
        for row in mdl.cache.df_in[mdl_cols].to_dict(orient='records'):
            adsys.add(mdl_name, row)

    _, s = elapsed(t0)

    # additonal file for dynamic models
    if addfile:
        t_add, _ = elapsed()

        # --- parse addfile ---
        adsys = parse_addfile(adsys=adsys, amsys=system, addfile=addfile)

        _, s_add = elapsed(t_add)
        logger.info('Addfile parsed in %s.', s_add)

    # fake FileManaer attributes
    adsys.files = system.files

    logger.info(f'System converted to ANDES in {s}.')

    # finalize
    system.dyn = Dynamic(amsys=system, adsys=adsys)
    system.dyn.link_andes(adsys=adsys)

    if setup:
        adsys.setup()
    elif verify:
        logger.warning('PFlow verification is skipped due to no setup.')
        return adsys
    if verify:
        verify_pf(amsys=system, adsys=adsys, tol=tol)
    return adsys


def parse_addfile(adsys, amsys, addfile):
    """
    Parse the addfile for ANDES dynamic file.

    Parameters
    ----------
    adsys : andes.system.System
        The ANDES system instance.
    amsys : ams.system.System
        The AMS system instance.
    addfile : str
        The additional file to be converted to ANDES dynamic mdoels.

    Returns
    -------
    adsys : andes.system.System
        The ANDES system instance with dynamic models added.
    """
    # guess addfile format
    add_format = None
    _, add_ext = os.path.splitext(addfile)
    for key, val in input_formats.items():
        if add_ext[1:] in val:
            add_format = key
            logger.debug('Addfile format guessed as %s.', key)
            break

    if key != 'xlsx':
        logger.error('Addfile format "%s" is not supported yet.', add_format)
        # FIXME: xlsx input file with dyr addfile result into KeyError: 'Toggle'
        # add_parser = importlib.import_module('andes.io.' + add_format)
        # if not add_parser.read_add(system, addfile):
        #     logger.error('Error parsing addfile "%s" with %s parser.', addfile, add_format)
        return adsys

    # Try parsing the addfile
    logger.info('Parsing additional file "%s"...', addfile)

    reader = pd.ExcelFile(addfile)

    pflow_mdl = list(pflow_dict.keys())

    pflow_mdls_overlap = []
    for mdl_name in pflow_dict.keys():
        if mdl_name in reader.sheet_names:
            pflow_mdls_overlap.append(mdl_name)

    if len(pflow_mdls_overlap) > 0:
        msg = 'Following PFlow models in addfile will be overwritten: '
        msg += ', '.join([f'<{mdl}>' for mdl in pflow_mdls_overlap])
        logger.warning(msg)

    pflow_mdl_nonempty = [mdl for mdl in pflow_mdl if amsys.models[mdl].n > 0]
    logger.debug(f"Non-empty PFlow models: {pflow_mdl}")
    pflow_df_models = pd.read_excel(addfile,
                                    sheet_name=pflow_mdl_nonempty,
                                    index_col=0,
                                    engine='openpyxl',
                                    )
    # drop rows that all nan
    for name, df in pflow_df_models.items():
        df.dropna(axis=0, how='all', inplace=True)

    # collect idx_map if difference exists
    idx_map = OrderedDict([])
    for name, df in pflow_df_models.items():
        am_idx = amsys.models[name].idx.v
        ad_idx = df['idx'].values
        if len(set(am_idx)) != len(set(ad_idx)):
            msg = f'<{name}> has different number of rows in addfile.'
            logger.warning(msg)
        if set(am_idx) != set(ad_idx):
            idx_map[name] = dict(zip(ad_idx, am_idx))

    # --- dynamic models to be added ---
    mdl_to_keep = list(set(reader.sheet_names) - set(pflow_mdl))
    mdl_to_keep.sort(key=str.lower)
    df_models = pd.read_excel(addfile,
                              sheet_name=mdl_to_keep,
                              index_col=0,
                              engine='openpyxl',
                              )

    # adjust models index
    for name, df in df_models.items():
        try:
            mdl = adsys.models[name]
        except KeyError:
            mdl = adsys.model_aliases[name]
        if len(mdl.idx_params) == 0:  # skip if no idx_params
            continue
        for idxn, idxp in mdl.idx_params.items():
            if idxp.model is None:  # make a guess if no model is specified
                mdl_guess = idxn.capitalize()
                if mdl_guess not in adsys.models.keys():
                    try:
                        mdl_guess = idx_guess[idxp.name]
                    except KeyError:  # set the most frequent string as the model name
                        split_list = []
                        for item in df[idxn].values:
                            if item is None or nan:
                                continue
                            try:
                                split_list.append(item.split('_'))
                                # Flatten the nested list and filter non-numerical strings
                                flattened_list = [item for sublist in split_list for item in sublist
                                                  if not isinstance(item, int)]
                                # Count the occurrences of non-numerical strings
                                string_counter = Counter(flattened_list)
                                # Find the most common non-numerical string
                                mdl_guess = string_counter.most_common(1)[0][0]
                            except AttributeError:
                                logger.error(f'Failed to parse IdxParam {name}.{idxn}.')
                                continue
            else:
                mdl_guess = idxp.model
            if mdl_guess in adsys.groups.keys():
                grp_idx = {}
                for mname, mdl in adsys.groups[mdl_guess].models.items():
                    # add group index to index map
                    if mname in idx_map.keys():
                        grp_idx.update(idx_map[mname])
                if len(grp_idx) == 0:
                    continue  # no index consistency issue, skip
                idx_map[mdl_guess] = grp_idx
            if mdl_guess not in idx_map.keys():
                continue  # no index consistency issue, skip
            else:
                logger.debug(f'Replace map for {mdl_guess} is {idx_map[mdl_guess]}')
                df[idxn] = df[idxn].replace(idx_map[mdl_guess])
                logger.debug(f'Adjust {idxp.class_name} <{name}.{idxp.name}>')

    # add dynamic models
    for name, df in df_models.items():
        # drop rows that all nan
        df.replace(['', ' '], nan, inplace=True)  # replace empty string with nan
        df.dropna(axis=0, how='all', inplace=True)
        # if the dynamic model also exists in AMS, use AMS parameters for overlap
        if (name in amsys.models.keys()) and amsys.models[name].n > 0:
            if df.shape[0] != amsys.models[name].n:
                msg = f'<{name}> has different number of rows in addfile.'
                logger.warning(msg)
            am_params = set(amsys.models[name].params.keys())
            ad_params = set(df.columns)
            overlap_params = list(am_params.intersection(ad_params))
            ad_rest_params = list(ad_params - am_params) + ['idx']
            msg = f'Following <{name}> parameters in addfile are overwriten: '
            msg += ', '.join(overlap_params)
            logger.debug(msg)
            tmp = amsys.models[name].cache.df_in[overlap_params]
            df = pd.merge(left=tmp, right=df[ad_rest_params],
                          on='idx', how='left')
        for row in df.to_dict(orient='records'):
            adsys.add(name, row)

    # --- adjust SynGen Vn with Bus Vn ---
    # NOTE: RenGen and DG have no Vn, so no need to adjust
    syg_idx = []
    for _, syg in adsys.SynGen.models.items():
        if syg.n > 0:
            syg_idx += syg.idx.v
    syg_bus_idx = adsys.SynGen.get(src='bus', attr='v', idx=syg_idx)
    syg_bus_vn = adsys.Bus.get(src='Vn', idx=syg_bus_idx)
    adsys.SynGen.set(src='Vn', idx=syg_idx, attr='v', value=syg_bus_vn)

    # --- for debugging ---
    adsys.df_in = df_models

    return adsys


class Dynamic:
    """
    ANDES interface class.

    Parameters
    ----------
    amsys : AMS.system.System
        The AMS system.
    adsys : ANDES.system.System
        The ANDES system.

    Attributes
    ----------
    link : pandas.DataFrame
        The ANDES system link table.

    Notes
    -----
    1. Using the file conversion ``to_andes()`` will automatically
       link the AMS system to the converted ANDES system in the
       attribute ``dyn``.

    Examples
    --------
    >>> import ams
    >>> import andes
    >>> sp = ams.load(ams.get_case('ieee14/ieee14_rted.xlsx'), setup=True)
    >>> sa = sp.to_andes(setup=True,
    ...                  addfile=andes.get_case('ieee14/ieee14_wt3.xlsx'),
    ...                  overwrite=True, keep=False, no_output=True)
    >>> sp.RTED.run()
    >>> sp.RTED.dc2ac()
    >>> sp.dyn.send()  # send RTED results to ANDES system
    >>> sa.PFlow.run()
    >>> sp.TDS.run()
    >>> sp.dyn.receive()  # receive TDS results from ANDES system
    """

    def __init__(self, amsys=None, adsys=None) -> None:
        self.amsys = amsys  # AMS system
        self.adsys = adsys  # ANDES system

        # TODO: add summary table
        self.link = None  # ANDES system link table

    def link_andes(self, adsys):
        """
        Link the ANDES system to the AMS system.

        Parameters
        ----------
        adsys : ANDES.system.System
            The ANDES system instance.
        """
        self.adsys = adsys

        self.link = make_link_table(self.adsys)
        logger.warning(f'AMS system {hex(id(self.amsys))} is linked to the ANDES system {hex(id(adsys))}.')

    @property
    def is_tds(self):
        """
        Indicator of whether the ANDES system is running a TDS.
        This property will return ``True`` as long as TDS is initialized.

        Check ``adsys.tds.TDS.init()`` for more details.
        """
        return bool(self.adsys.TDS.initialized)

    def _send_tgr(self, sa, sp):
        """
        Sned to generator power refrence.

        Notes
        -----
        1. AGC power reference ``paux`` is not included in this function.
        """
        # 1) TurbineGov
        syg_idx = sp.dyn.link['syg_idx'].dropna().tolist()  # SynGen idx
        # corresponding StaticGen idx in ANDES
        stg_syg_idx = sa.SynGen.get(src='gen', attr='v', idx=syg_idx,
                                    allow_none=True, default=None)
        # corresponding TurbineGov idx in ANDES
        gov_idx = sa.TurbineGov.find_idx(keys='syn', values=syg_idx)
        # corresponding StaticGen pg in AMS
        syg_ams = sp.recent.get(src='pg', attr='v', idx=stg_syg_idx,
                                allow_none=True, default=0)
        # --- check consistency ---
        syg_mask = self.link['syg_idx'].notnull() & self.link['gov_idx'].isnull()
        if syg_mask.any():
            logger.debug('Governor is not complete for SynGen.')
        # --- pref ---
        sa.TurbineGov.set(value=syg_ams, idx=gov_idx, attr='v', src='pref0')

        # --- paux ---
        # TODO: sync paux, using paux0

        # 2) DG
        dg_idx = sp.dyn.link['dg_idx'].dropna().tolist()  # DG idx
        # corresponding StaticGen idx in ANDES
        stg_dg_idx = sa.DG.get(src='gen', attr='v', idx=dg_idx,
                               allow_none=True, default=None,
                               )
        # corresponding StaticGen pg in AMS
        dg_ams = sp.recent.get(src='pg', attr='v', idx=stg_dg_idx,
                               allow_none=True, default=0)
        # --- pref ---
        sa.DG.set(src='pref0', idx=dg_idx, attr='v', value=dg_ams)
        # TODO: paux, using Pext0, this one should be do in other place rather than here

        # 3) RenGen
        # TODO: seems to be unnecessary
        # which models/params are used to control output and auxillary power?

        return True

    def _send_dgu(self, sa, sp):
        """
        Send to ANDES the dynamic generator online status.
        """
        # 1) SynGen
        syg_idx = sp.dyn.link['syg_idx'].dropna().tolist()  # SynGen idx
        # corresponding StaticGen idx in ANDES
        stg_syg_idx = sa.SynGen.get(src='gen', attr='v', idx=syg_idx,
                                    allow_none=True, default=None)
        # corresponding StaticGen u in AMS
        stg_u_ams = sp.StaticGen.get(src='u', attr='v', idx=stg_syg_idx,
                                     allow_none=True, default=0)
        stg_u_andes = sa.SynGen.get(src='u', attr='v', idx=syg_idx,
                                    allow_none=True, default=0)
        # 2) DG
        dg_idx = sp.dyn.link['dg_idx'].dropna().tolist()  # DG idx
        # corresponding StaticGen idx in ANDES
        stg_dg_idx = sa.DG.get(src='gen', attr='v', idx=dg_idx,
                               allow_none=True, default=None)
        # corresponding DG u in AMS
        dg_u_ams = sp.StaticGen.get(src='u', attr='v', idx=stg_dg_idx,
                                    allow_none=True, default=0)
        du_u_andes = sa.DG.get(src='u', attr='v', idx=dg_idx,
                               allow_none=True, default=0)
        # 3) RenGen
        rg_idx = sp.dyn.link['rg_idx'].dropna().tolist()  # RenGen idx
        # corresponding StaticGen idx in ANDES
        stg_rg_idx = sa.RenGen.get(src='gen', attr='v', idx=rg_idx,
                                   allow_none=True, default=None)
        # corresponding RenGen u in AMS
        rg_u_ams = sp.StaticGen.get(src='u', attr='v', idx=stg_rg_idx,
                                    allow_none=True, default=0)
        rg_u_andes = sa.RenGen.get(src='u', attr='v', idx=rg_idx,
                                   allow_none=True, default=0)
        # 4) sync results
        cond = (
            not np.array_equal(stg_u_ams, stg_u_andes) or
            not np.array_equal(dg_u_ams, du_u_andes) or
            not np.array_equal(rg_u_ams, rg_u_andes)
        )
        if cond:
            msg = 'ANDES dynamic generator online status should be switched using Toggle!'
            msg += ' Otherwise, unexpected results might occur.'
            raise ValueError(msg)
        # FIXME: below code seems to be unnecessary
        sa.SynGen.set(src='u', idx=syg_idx, attr='v', value=stg_u_ams)
        sa.DG.set(src='u', idx=dg_idx, attr='v', value=dg_u_ams)
        sa.RenGen.set(src='u', idx=rg_idx, attr='v', value=rg_u_ams)
        return True

    def _sync_check(self, amsys, adsys):
        """
        Check if AMS and ANDES systems are ready for sync.
        """
        if amsys.dyn.adsys:
            if amsys.dyn.adsys != adsys:
                logger.error('Target ANDES system is different from the linked one, quit.')
                return False
        if not amsys.is_setup:
            amsys.setup()
        if not adsys.is_setup:
            adsys.setup()
        if amsys.dyn.link is None:
            amsys.dyn.link = make_link_table(adsys=adsys)

    def send(self, adsys=None, routine=None):
        """
        Send results of the recent sovled AMS routine (``sp.recent``) to the
        target ANDES system.

        Note that converged AC conversion DOES NOT guarantee successful dynamic
        initialization ``TDS.init()``.
        Failed initialization is usually caused by limiter violation.

        Parameters
        ----------
        adsys : adsys.System.system, optional
            The target ANDES dynamic system instance. If not provided, use the
            linked ANDES system isntance (``sp.dyn.adsys``).
        routine : str, optional
            The routine to be sent to ANDES. If None, ``recent`` will be used.
        """
        sa = adsys if adsys is not None else self.adsys
        sp = self.amsys
        self._sync_check(amsys=sp, adsys=sa)

        # --- Information ---
        rtn = sp.recent if routine is None else getattr(sp, routine)
        if rtn is None:
            logger.warning('No assigned or recent solved routine found, quit send.')
            return False
        elif rtn.exit_code != 0:
            logger.warning(f'{sp.recent.class_name} is not solved at optimal, quit send.')
            return False
        else:
            logger.info(f'Send <{rtn.class_name}> results to ANDES <{hex(id(sa))}>...')

        # NOTE: if DC type, check if results are converted
        if (rtn.type != 'ACED') and (not rtn.converted):
            logger.error(f'<{rtn.class_name}> AC conversion failed or not done yet!')

        # --- Mapping ---
        map2 = getattr(rtn, 'map2')     # mapping-to dict
        if len(map2) == 0:
            logger.warning(f'{rtn.class_name} has empty map2, quit send.')
            return True

        # NOTE: ads is short for ANDES
        for vname_ams, (mname_ads, pname_ads) in map2.items():
            mdl_ads = getattr(sa, mname_ads)  # ANDES model or group

            # --- skipping scenarios ---
            if mdl_ads.n == 0:
                logger.debug(f'ANDES model <{mname_ads}> is empty.')
                continue

            var_ams = getattr(rtn, vname_ams)  # instance of AMS routine var
            idx_ads = var_ams.get_idx()  # use AMS idx as target ANDES idx

            # --- special scenarios ---
            # 0. send PV bus voltage to StaticGen.v0 if not PFlow yet and AC converted
            cond_vpv = (mname_ads == 'Bus') and (pname_ads == 'v0')
            if cond_vpv and (not self.is_tds) and (rtn.converted):
                # --- StaticGen ---
                stg_idx = sp.StaticGen.get_idx()
                bus_stg = sp.StaticGen.get(src='bus', attr='v', idx=stg_idx)
                vBus = rtn.get(src='vBus', attr='v', idx=bus_stg)
                sa.StaticGen.set(src='v0', idx=stg_idx, attr='v', value=vBus)
                logger.info(f'*Send <{vname_ams}> to StaticGen.v0')

            # 1. gen online status; in TDS running, setting u is invalid
            cond_ads_stg_u = (mname_ads in ['StaticGen', 'PV', 'Sclak']) and (pname_ads == 'u')
            if cond_ads_stg_u and (self.is_tds):
                logger.info(f'*Skip sending {vname_ams} to StaticGen.u during TDS')
                continue

            # 2. Bus voltage
            cond_ads_bus_v0 = (mname_ads == 'Bus') and (pname_ads == 'v0')
            if cond_ads_bus_v0 and (self.is_tds):
                logger.info(f'*Skip sending {vname_ams} t0 Bus.v0 during TDS')
                continue

            # 3. gen power reference; in TDS running, pg should go to TurbineGov
            cond_ads_stg_p0 = (mname_ads in ['StaticGen', 'PV', 'Sclak']) and (pname_ads == 'p0')
            if cond_ads_stg_p0 and (self.is_tds):
                # --- SynGen: TurbineGov.pref0 ---
                syg_idx = sp.dyn.link['syg_idx'].dropna().tolist()  # SynGen idx
                # corresponding StaticGen idx in ANDES
                stg_syg_idx = sa.SynGen.get(src='gen', attr='v', idx=syg_idx,
                                            allow_none=True, default=None)
                # corresponding TurbineGov idx in ANDES
                gov_idx = sa.TurbineGov.find_idx(keys='syn', values=syg_idx)
                # corresponding StaticGen pg in AMS
                syg_ams = rtn.get(src='pg', attr='v', idx=stg_syg_idx)
                # NOTE: check consistency
                syg_mask = self.link['syg_idx'].notnull() & self.link['gov_idx'].isnull()
                if syg_mask.any():
                    logger.debug('Governor is not complete for SynGen.')
                # --- pref ---
                sa.TurbineGov.set(src='pref0', idx=gov_idx, attr='v', value=syg_ams)

                # --- DG: DG.pref0 ---
                dg_idx = sp.dyn.link['dg_idx'].dropna().tolist()  # DG idx
                # corresponding StaticGen idx in ANDES
                stg_dg_idx = sa.DG.get(src='gen', attr='v', idx=dg_idx,
                                       allow_none=True, default=None)
                # corresponding StaticGen pg in AMS
                dg_ams = rtn.get(src='pg', attr='v', idx=stg_dg_idx)
                # --- pref ---
                sa.DG.set(src='pref0', idx=dg_idx, attr='v', value=dg_ams)

                # --- RenGen: seems unnecessary ---
                # TODO: which models/params are used to control output and auxillary power?

                var_dest = ''
                if len(syg_ams) > 0:
                    var_dest = 'TurbineGov.pref0'
                if len(dg_ams) > 0:
                    var_dest += ' and DG.pref0'
                logger.warning(f'*Send <{vname_ams}> to {var_dest}')
                continue

            # --- other scenarios ---
            if _dest_check(mname=mname_ads, pname=pname_ads, idx=idx_ads, adsys=sa):
                mdl_ads.set(src=pname_ads, idx=idx_ads, attr='v', value=var_ams.v)
                logger.warning(f'Send <{vname_ams}> to {mname_ads}.{pname_ads}')
        return True

    def receive(self, adsys=None, routine=None, no_update=False):
        """
        Receive ANDES system results to AMS devices.

        Parameters
        ----------
        adsys : adsys.System.system, optional
            The target ANDES dynamic system instance. If not provided, use the
            linked ANDES system isntance (``sp.dyn.adsys``).
        routine : str, optional
            The routine to be received from ANDES. If None, ``recent`` will be used.
        no_update : bool, optional
            True to skip update the AMS routine parameters after sync. Default is False.
        """
        sa = adsys if adsys is not None else self.adsys
        sp = self.amsys
        self._sync_check(amsys=sp, adsys=sa)

        #  --- Information: not necessary in receive ---
        rtn = sp.recent if routine is None else getattr(sp, routine)
        if rtn is None:
            logger.warning('No assigned or recent solved routine found, quit receive.')
            return False

        # --- Mapping ---
        map1 = getattr(rtn, 'map1')     # mapping-from dict
        if len(map1) == 0:
            logger.warning(f'{rtn.class_name} has empty map1, quit receive.')
            return True

        link = sp.dyn.link      # link table
        pname_to_update = []    # list of parameters to be updated
        for vname_ams, (mname_ads, pname_ads) in map1.items():
            mdl_ads = getattr(sa, mname_ads)  # ANDES model or group

            # --- skipping scenarios ---
            if mdl_ads.n == 0:
                logger.debug(f'ANDES model <{mname_ads}> is empty.')
                continue

            idx_ads = rtn.__dict__[vname_ams].get_idx()  # use AMS idx as target ANDES idx

            # --- special scenarios ---
            # 1. gen online status; in TDS running, take from dynamic generator
            cond_ads_stg_u = (mname_ads in ['StaticGen', 'PV', 'Sclak']) and (pname_ads == 'u')
            if cond_ads_stg_u and (self.is_tds):
                # --- SynGen ---
                u_sg = sa.SynGen.get(idx=link['syg_idx'].replace(nan, None).to_list(),
                                     src='u', attr='v',
                                     allow_none=True, default=0,)
                # --- DG ---
                u_dg = sa.DG.get(idx=link['dg_idx'].replace(nan, None).to_list(),
                                 src='u', attr='v',
                                 allow_none=True, default=0,)
                # --- RenGen ---
                u_rg = sa.RenGen.get(idx=link['rg_idx'].replace(nan, None).to_list(),
                                     src='u', attr='v',
                                     allow_none=True, default=0,)
                # --- output ---
                u_dyg = u_sg + u_rg + u_dg
                # NOTE: StaticGen that has no dynamic generator
                # Sync StaticGen.u first, then overwrite the ones with dynamic generator
                u_stg = sa.StaticGen.get(src='u', attr='v',
                                         idx=link['stg_idx'].values)
                # NOTE: only update u if changed actually
                u0_rtn = rtn.get(src=vname_ams, attr='v', idx=link['stg_idx'].values).copy()
                rtn.set(src=vname_ams, idx=link['stg_idx'].values, attr='v', value=u_stg)
                rtn.set(src=vname_ams, idx=link['stg_idx'].values, attr='v', value=u_dyg)
                u_rtn = rtn.get(src=vname_ams, attr='v', idx=link['stg_idx'].values).copy()
                if not np.array_equal(u0_rtn, u_rtn):
                    pname_to_update.append(vname_ams)
                var_dest = ''
                var_dest += 'SynGen.u' if sa.SynGen.n > 0 else ''
                var_dest += ', DG.u' if sa.DG.n > 0 else ''
                var_dest += ', RenGen.u' if sa.RenGen.n > 0 else ''
                var_dest += ', StaicGen.u' if sa.RenGen.n+sa.SynGen.n+sa.DG.n < sa.StaticGen.n else ''
                logger.info(f'Receive <{vname_ams}> from {var_dest}')
                continue

            # 2. gen output power; in TDS running, take from dynamic generator
            cond_ads_stg_p = (mname_ads in ['StaticGen', 'PV', 'Sclak']) and (pname_ads == 'p')
            if cond_ads_stg_p and (self.is_tds):
                # --- SynGen ---
                p_sg = sa.SynGen.get(idx=link['syg_idx'].replace(nan, None).to_list(),
                                     src='Pe', attr='v',
                                     allow_none=True, default=0,)
                # --- DG ---
                Ie_dg = sa.DG.get(idx=link['dg_idx'].replace(nan, None).to_list(),
                                  src='Ipout_y', attr='v',
                                  allow_none=True, default=0,)
                v_dg = sa.DG.get(idx=link['dg_idx'].replace(nan, None).to_list(),
                                 src='v', attr='v',
                                 allow_none=True, default=0,)
                p_dg = Ie_dg * v_dg
                # --- RenGen ---
                p_rg = sa.RenGen.get(idx=link['rg_idx'].replace(nan, None).to_list(),
                                     src='Pe', attr='v',
                                     allow_none=True, default=0,)
                # --- output ---
                p_dyg = p_sg + p_rg + p_dg
                # NOTE: StaticGen that has no dynamic generator
                # Sync StaticGen.p first, then overwrite the ones with dynamic generator
                p_stg = sa.StaticGen.get(src='p', attr='v',
                                         idx=link['stg_idx'].values)
                rtn.set(src=vname_ams, idx=link['stg_idx'].values, attr='v', value=p_stg)
                rtn.set(src=vname_ams, idx=link['stg_idx'].values, attr='v', value=p_dyg)

                pname_to_update.append(vname_ams)

                var_dest = ''
                var_dest += 'SynGen.Pe' if sa.SynGen.n > 0 else ''
                var_dest += ', DG.Pe' if sa.DG.n > 0 else ''
                var_dest += ', RenGen.Pe' if sa.RenGen.n > 0 else ''
                var_dest += ', StaicGen.p' if sa.RenGen.n+sa.SynGen.n+sa.DG.n < sa.StaticGen.n else ''
                logger.info(f'Receive <{vname_ams}> from {var_dest}')
                continue

            # --- other scenarios ---
            if _dest_check(mname=mname_ads, pname=pname_ads, idx=idx_ads, adsys=sa):
                v_ads = mdl_ads.get(src=pname_ads, attr='v', idx=idx_ads)
                rtn.set(src=vname_ams, idx=idx_ads, attr='v', value=v_ads)
                pname_to_update.append(vname_ams)
                logger.warning(f'Receive <{vname_ams}> from {mname_ads}.{pname_ads}')

        # --- update routine parameters ---
        if no_update and (len(pname_to_update) > 0):
            logger.info(f'Please update <{rtn.class_name}> parameters: {pname_to_update}')
        elif len(pname_to_update) > 0:
            rtn.update(params=pname_to_update, build_mats=False)
        return True


def _dest_check(mname, pname, idx, adsys):
    """
    Check if destination is valid.

    Parameters
    ----------
    mname : str
        Target ANDES model/group name.
    pname : str
        Target ANDES parameter name.
    idx : list
        Target idx.
    adsys : ANDES.system.System
        Target ANDES system.
    """
    # --- check model ---
    if not hasattr(adsys, mname):
        raise ValueError(f'Model error: ANDES system <{hex(adsys)}> has no <{mname}>')

    # --- check param ---
    mdl = getattr(adsys, mname)
    _is_grp = mname in adsys.groups.keys()
    # if it is a group, iterate through all models in the group
    mdls_grp_name = list(adsys.groups[mname].models.keys()) if _is_grp else [mname]
    n_no_mdl = 0
    for mdl_name in mdls_grp_name:
        mdl_to_check = getattr(adsys, mdl_name, None)
        if mdl_to_check is not None and hasattr(mdl_to_check, pname):
            break   # found the param in any one of the models, break
        n_no_mdl += 1
    if n_no_mdl == len(mdls_grp_name):
        raise ValueError(f'Param error: ANDES <{mdl.class_name}> has no <{pname}>')

    # --- check idx ---
    if _is_grp:
        _ads_idx = [v for mdl in mdl.models.values() for v in mdl.idx.v]
    else:
        _ads_idx = mdl.idx.v
    if not set(idx).issubset(set(_ads_idx)):
        idx_gap = set(idx) - set(_ads_idx)
        raise ValueError(f'Idx error: ANDES <{mdl.class_name}> has no idx: {idx_gap}')

    return True


def build_group_table(adsys, grp_name, param_name, mdl_name=None):
    """
    Build the table for devices in a group in an ANDES System.

    Parameters
    ----------
    adsys : andes.system.System
        The ANDES system to build the table
    grp_name : string
        The ANDES group
    param_name : list of string
        The common columns of a group that to be included in the table.
    mdl_name : list of string
        The list of models that to be included in the table. Default as all models.

    Returns
    -------
    DataFrame

        The output Dataframe contains the columns from the device
    """
    grp_df = pd.DataFrame(columns=param_name)
    grp = getattr(adsys, grp_name)  # get the group instance

    mdl_to_add = mdl_name if mdl_name else list(grp.models.keys())
    mdl_dfs = [getattr(adsys, mdl).as_df()[param_name] for mdl in mdl_to_add]

    grp_df = pd.concat(mdl_dfs, axis=0, ignore_index=True)

    # --- type sanity check ---
    mdl_1st = adsys.models[mdl_to_add[0]]
    # NOTE: force IdxParam to be string type
    cols_to_convert = [col for col in param_name if
                       (mdl_1st.params[col].class_name == 'IdxParam') and
                       (pd.api.types.is_numeric_dtype(grp_df[col]))]
    # NOTE: if 'idx' is included, force it to be string type
    if ('idx' in param_name) and pd.api.types.is_numeric_dtype(grp_df['idx']):
        cols_to_convert.append('idx')

    grp_df[cols_to_convert] = grp_df[cols_to_convert].astype(int).astype(str)
    return grp_df


def make_link_table(adsys):
    """
    Build the link table for generators and generator controllers in an ANDES
    System, including ``SynGen`` and ``DG`` for now.

    Parameters
    ----------
    adsys : andes.system.System
        The ANDES system to link

    Returns
    -------
    DataFrame

        Each column in the output Dataframe contains the ``idx`` of linked
        ``StaticGen``, ``Bus``, ``DG``, ``RenGen``, ``RenExciter``, ``SynGen``,
        ``Exciter``, and ``TurbineGov``, ``gammap``, ``gammaq``.
    """
    # --- build group tables ---
    # 1) StaticGen
    ssa_stg = build_group_table(adsys=adsys, grp_name='StaticGen',
                                param_name=['u', 'name', 'idx', 'bus'],
                                mdl_name=[])
    # 2) TurbineGov
    ssa_gov = build_group_table(adsys=adsys, grp_name='TurbineGov',
                                param_name=['idx', 'syn'],
                                mdl_name=[])
    # 3) Exciter
    ssa_exc = build_group_table(adsys=adsys, grp_name='Exciter',
                                param_name=['idx', 'syn'],
                                mdl_name=[])
    # 4) SynGen
    ssa_syg = build_group_table(adsys=adsys, grp_name='SynGen', mdl_name=['GENCLS', 'GENROU'],
                                param_name=['idx', 'bus', 'gen', 'gammap', 'gammaq'])
    # 5) DG
    ssa_dg = build_group_table(adsys=adsys, grp_name='DG', mdl_name=[],
                               param_name=['idx', 'bus', 'gen', 'gammap', 'gammaq'])
    # 6) RenGen
    ssa_rg = build_group_table(adsys=adsys, grp_name='RenGen', mdl_name=[],
                               param_name=['idx', 'bus', 'gen', 'gammap', 'gammaq'])
    # 7) RenExciter
    ssa_rexc = build_group_table(adsys=adsys, grp_name='RenExciter', mdl_name=[],
                                 param_name=['idx', 'reg'])

    # --- build link table ---
    # NOTE: use bus index as unique identifier
    ssa_bus = build_group_table(adsys=adsys, grp_name='ACTopology', mdl_name=['Bus'],
                                param_name=['name', 'idx'])
    # 1) StaticGen
    ssa_key = pd.merge(left=ssa_stg.rename(columns={'name': 'stg_name', 'idx': 'stg_idx',
                                                    'bus': 'bus_idx', 'u': 'stg_u'}),
                       right=ssa_bus.rename(columns={'name': 'bus_name', 'idx': 'bus_idx'}),
                       how='left', on='bus_idx')
    # 2) Dynamic Generators
    ssa_syg = pd.merge(left=ssa_key, how='right', on='stg_idx',
                       right=ssa_syg.rename(columns={'idx': 'syg_idx', 'gen': 'stg_idx'}))
    ssa_dg = pd.merge(left=ssa_key, how='right', on='stg_idx',
                      right=ssa_dg.rename(columns={'idx': 'dg_idx', 'gen': 'stg_idx'}))
    ssa_rg = pd.merge(left=ssa_key, how='right', on='stg_idx',
                      right=ssa_rg.rename(columns={'idx': 'rg_idx', 'gen': 'stg_idx'}))

    # NOTE: for StaticGen without Dynamic Generator, fill gammap and gammaq as 1
    ssa_key0 = pd.merge(left=ssa_key, how='left', on='stg_idx',
                        right=ssa_syg[['stg_idx', 'syg_idx']])
    ssa_key0 = pd.merge(left=ssa_key0, how='left', on='stg_idx',
                        right=ssa_dg[['stg_idx', 'dg_idx']])
    ssa_key0 = pd.merge(left=ssa_key0, how='left', on='stg_idx',
                        right=ssa_rg[['stg_idx', 'rg_idx']])

    # NOTE: use this instead of fillna to avoid type conversion
    idxc = ['stg_idx', 'syg_idx', 'dg_idx', 'rg_idx']
    ssa_key0[idxc] = ssa_key0[idxc].astype('str').replace({'nan': ''}).astype('bool')

    dyr = ssa_key0['syg_idx'] + ssa_key0['dg_idx'] + ssa_key0['rg_idx']
    non_dyr = np.logical_not(dyr)
    ssa_dyr0 = ssa_key0[non_dyr]
    ssa_dyr0['gammap'] = 1
    ssa_dyr0['gammaq'] = 1

    ssa_key = pd.concat([ssa_syg, ssa_dg, ssa_rg, ssa_dyr0], axis=0)
    ssa_key = pd.merge(left=ssa_key,
                       right=ssa_exc.rename(columns={'idx': 'exc_idx', 'syn': 'syg_idx'}),
                       how='left', on='syg_idx')
    ssa_key = pd.merge(left=ssa_key,
                       right=ssa_gov.rename(columns={'idx': 'gov_idx', 'syn': 'syg_idx'}),
                       how='left', on='syg_idx')
    ssa_key = pd.merge(left=ssa_key, how='left', on='rg_idx',
                       right=ssa_rexc.rename(columns={'idx': 'rexc_idx', 'reg': 'rg_idx'}))

    # NOTE: other cols might be useful in the future
    # cols = ['stg_name', 'stg_u', 'stg_idx', 'bus_idx', 'dg_idx', 'rg_idx', 'rexc_idx',
    #         'syg_idx', 'exc_idx', 'gov_idx', 'bus_name', 'gammap', 'gammaq']
    # re-order columns
    cols = ['stg_idx', 'bus_idx',               # static gen
            'syg_idx', 'gov_idx',               # syn gen
            'dg_idx',                           # distributed gen
            'rg_idx',                           # renewable gen
            'gammap', 'gammaq',                 # gamma
            ]
    out = ssa_key[cols].sort_values(by='stg_idx', ascending=False).reset_index(drop=True)
    return out


def verify_pf(amsys, adsys, tol=1e-3):
    """
    Verify the power flow results between AMS and ANDES.
    Note that this function will run PFlow in both systems.

    Parameters
    ----------
    sp : ams.System
        The AMS system.
    sa : andes.System
        The ANDES system.

    Returns
    -------
    bool
        True if the power flow results are consistent; False otherwise.
    """
    amsys.PFlow.run()
    if adsys.is_setup:
        adsys.PFlow.run()
    else:
        logger.info('ANDES system is not setup, quit verification.')
        return False
    v_check = np.allclose(amsys.Bus.v.v, adsys.Bus.v.v, atol=tol)
    a_check = np.allclose(amsys.Bus.a.v, adsys.Bus.a.v, atol=tol)
    check = v_check and a_check

    v_diff_max = np.max(np.abs(amsys.Bus.v.v - adsys.Bus.v.v))
    a_diff_max = np.max(np.abs(amsys.Bus.a.v - adsys.Bus.a.v))
    diff_msg = f'Voltage diff max: {v_diff_max}, Angle diff max: {a_diff_max}'
    logger.debug(diff_msg)
    if check:
        logger.info('Power flow results are consistent.')
    else:
        msg = 'Power flow results are inconsistent!'
        logger.warning(msg)
        logger.warning(diff_msg)
    return check



"""
Test ANDES matrices.
"""

import unittest
import numpy as np
import importlib.metadata
from packaging.version import parse as parse_version

import ams


class TestMatrices(unittest.TestCase):
    """
    Tests for system matrices consistency.
    """

    andes_version = importlib.metadata.version("andes")
    if parse_version(andes_version) < parse_version('1.9.2'):
        raise unittest.SkipTest("Requires ANDES version >= 1.9.2")

    sp = ams.load(ams.get_case('5bus/pjm5bus_demo.xlsx'),
                  setup=True, no_output=True, default_config=True,)
    sa = sp.to_andes(setup=True, no_output=True, default_config=True,)

    def setUp(self) -> None:
        """
        Test setup.
        """

    def test_build_y(self):
        """
        Test build_y consistency.
        """
        ysp = self.sp.Line.build_y()
        ysa = self.sa.Line.build_y()
        np.testing.assert_equal(np.array(ysp.V), np.array(ysa.V))

    def test_build_Bp(self):
        """
        Test build_Bp consistency.
        """
        Bp_sp = self.sp.Line.build_Bp()
        Bp_sa = self.sa.Line.build_Bp()
        np.testing.assert_equal(np.array(Bp_sp.V), np.array(Bp_sa.V))

    def test_build_Bpp(self):
        """
        Test build_Bpp consistency.
        """
        Bpp_sp = self.sp.Line.build_Bpp()
        Bpp_sa = self.sa.Line.build_Bpp()
        np.testing.assert_equal(np.array(Bpp_sp.V), np.array(Bpp_sa.V))

    def test_build_Bdc(self):
        """
        Test build_Bdc consistency.
        """
        Bdc_sp = self.sp.Line.build_Bdc()
        Bdc_sa = self.sa.Line.build_Bdc()
        np.testing.assert_equal(np.array(Bdc_sp.V), np.array(Bdc_sa.V))



"""
This file is used to generate reStructuredText tables for Model and Routine references.
"""

import os
import ams

if not (os.path.isfile('routineref.rst') and os.path.isfile('configref.rst')):

    ss = ams.load(ams.get_case('ieee14/ieee14_uced.xlsx'))

    # write the top-level index file

    out = """.. _routineref:

*******************
Routine reference
*******************

Use the left navigation pane to locate the group and model
and view details.

"""

    out += ss.supported_routines(export='rest')

    out += '\n'
    out += '.. toctree ::\n'
    out += '    :maxdepth: 2\n'
    out += '    :hidden:\n'
    out += '\n'

    file_tpl = '    typedoc/{}\n'

    for rtn_type in ss.types.values():
        out += file_tpl.format(rtn_type.class_name)

    with open('routineref.rst', 'w') as f:
        f.write(out)

    # write individual files

    os.makedirs('typedoc', exist_ok=True)

    for rtn_type in ss.types.values():
        with open(f'typedoc/{rtn_type.class_name}.rst', 'w') as f:
            f.write(rtn_type.doc_all(export='rest'))



"""
Module for optimization modeling.
"""
import logging

from typing import Any, Optional, Union
from collections import OrderedDict
import re

import numpy as np

from andes.core.common import Config
from andes.utils.misc import elapsed

import cvxpy as cp

from ams.utils import pretty_long_message
from ams.shared import sps

logger = logging.getLogger(__name__)

_prefix = r" - --------------> | "
_max_length = 80


def ensure_symbols(func):
    """
    Decorator to ensure that symbols are generated before parsing.
    If not, it runs self.rtn.syms.generate_symbols().

    Designed to be used on the `parse` method of the optimization elements (`OptzBase`)
    and optimization model (`OModel`), i.e., `Var`, `Param`, `Constraint`, `Objective`,
    and `ExpressionCalc`.

    Note:
    -----
    Parsing before symbol generation can give wrong results. Ensure that symbols
    are generated before calling the `parse` method.
    """

    def wrapper(self, *args, **kwargs):
        if not self.rtn._syms:
            logger.debug(f"<{self.rtn.class_name}> symbols are not generated yet. Generating now...")
            self.rtn.syms.generate_symbols()
        return func(self, *args, **kwargs)
    return wrapper


def ensure_mats_and_parsed(func):
    """
    Decorator to ensure that system matrices are built and OModel is parsed
    before evaluation. If not, it runs the necessary methods to initialize them.

    Designed to be used on the `evaluate` method of the optimization elements (`OptzBase`)
    and optimization model (`OModel`), i.e., `Var`, `Param`, `Constraint`, `Objective`,
    and `ExpressionCalc`.

    Note:
    -----
    Evaluation before matrices building and parsing can run into errors. Ensure that
    system matrices are built and OModel is parsed before calling the `evaluate` method.
    """

    def wrapper(self, *args, **kwargs):
        try:
            if not self.rtn.system.mats.initialized:
                logger.debug("System matrices are not built yet. Building now...")
                self.rtn.system.mats.build()
            if isinstance(self, (OptzBase, Var, Param, Constraint, Objective)):
                if not self.om.parsed:
                    logger.debug("OModel is not parsed yet. Parsing now...")
                    self.om.parse()
            elif isinstance(self, OModel):
                if not self.parsed:
                    logger.debug("OModel is not parsed yet. Parsing now...")
                    self.parse()
        except Exception as e:
            logger.error(f"Error during initialization or parsing: {e}")
            raise
        return func(self, *args, **kwargs)
    return wrapper


class OptzBase:
    """
    Base class for optimization elements, e.g., Var and Constraint.

    Parameters
    ----------
    name : str, optional
        Name.
    info : str, optional
        Descriptive information

    Attributes
    ----------
    rtn : ams.routines.Routine
        The owner routine instance.

    Note:
    -----
    Ensure that symbols are generated before calling the `parse` method. Parsing
    before symbol generation can give wrong results.
    """

    def __init__(self,
                 name: Optional[str] = None,
                 info: Optional[str] = None,
                 unit: Optional[str] = None,
                 ):
        self.om = None
        self.name = name
        self.info = info
        self.unit = unit
        self.is_disabled = False
        self.rtn = None
        self.optz = None  # corresponding optimization element
        self.code = None

    @ensure_symbols
    def parse(self):
        """
        Parse the object.
        """
        raise NotImplementedError

    @ensure_mats_and_parsed
    def evaluate(self):
        """
        Evaluate the object.
        """
        raise NotImplementedError

    @property
    def class_name(self):
        """
        Return the class name
        """
        return self.__class__.__name__

    @property
    def n(self):
        """
        Return the number of elements.
        """
        if self.owner is None:
            return len(self.v)
        else:
            return self.owner.n

    @property
    def shape(self):
        """
        Return the shape.
        """
        try:
            return self.om.__dict__[self.name].shape
        except KeyError:
            logger.warning('Shape info is not ready before initialization.')
            return None

    @property
    def size(self):
        """
        Return the size.
        """
        if self.rtn.initialized:
            return self.om.__dict__[self.name].size
        else:
            logger.warning(f'Routine <{self.rtn.class_name}> is not initialized yet.')
            return None


class ExpressionCalc(OptzBase):
    """
    Expression for calculation.
    """

    def __init__(self,
                 name: Optional[str] = None,
                 info: Optional[str] = None,
                 unit: Optional[str] = None,
                 var: Optional[str] = None,
                 e_str: Optional[str] = None,
                 ):
        OptzBase.__init__(self, name=name, info=info, unit=unit)
        self.optz = None
        self.var = var
        self.e_str = e_str
        self.code = None

    @ensure_symbols
    def parse(self):
        """
        Parse the Expression.
        """
        # parse the expression str
        sub_map = self.om.rtn.syms.sub_map
        code_expr = self.e_str
        for pattern, replacement in sub_map.items():
            try:
                code_expr = re.sub(pattern, replacement, code_expr)
            except Exception as e:
                raise Exception(f"Error in parsing expr <{self.name}>.\n{e}")
        # store the parsed expression str code
        self.code = code_expr
        msg = f" - ExpressionCalc <{self.name}>: {self.e_str}"
        logger.debug(pretty_long_message(msg, _prefix, max_length=_max_length))
        return True

    @ensure_mats_and_parsed
    def evaluate(self):
        """
        Evaluate the expression.
        """
        msg = f" - Expression <{self.name}>: {self.code}"
        logger.debug(pretty_long_message(msg, _prefix, max_length=_max_length))
        try:
            local_vars = {'self': self, 'np': np, 'cp': cp}
            self.optz = self._evaluate_expression(self.code, local_vars=local_vars)
        except Exception as e:
            raise Exception(f"Error in evaluating expr <{self.name}>.\n{e}")
        return True

    def _evaluate_expression(self, code, local_vars=None):
        """
        Helper method to evaluate the expression code.

        Parameters
        ----------
        code : str
            The code string representing the expression.

        Returns
        -------
        cp.Expression
            The evaluated cvxpy expression.
        """
        return eval(code, {}, local_vars)

    @property
    def v(self):
        """
        Return the CVXPY expression value.
        """
        if self.optz is None:
            return None
        else:
            return self.optz.value

    def __repr__(self):
        return f'{self.__class__.__name__}: {self.name}'


class Param(OptzBase):
    """
    Base class for parameters used in a routine.

    Parameters
    ----------
    no_parse: bool, optional
        True to skip parsing the parameter.
    nonneg: bool, optional
        True to set the parameter as non-negative.
    nonpos: bool, optional
        True to set the parameter as non-positive.
    cplx: bool, optional
        True to set the parameter as complex, avoiding the use of `complex`.
    imag: bool, optional
        True to set the parameter as imaginary.
    symmetric: bool, optional
        True to set the parameter as symmetric.
    diag: bool, optional
        True to set the parameter as diagonal.
    hermitian: bool, optional
        True to set the parameter as hermitian.
    boolean: bool, optional
        True to set the parameter as boolean.
    integer: bool, optional
        True to set the parameter as integer.
    pos: bool, optional
        True to set the parameter as positive.
    neg: bool, optional
        True to set the parameter as negative.
    sparse: bool, optional
        True to set the parameter as sparse.
    """

    def __init__(self,
                 name: Optional[str] = None,
                 info: Optional[str] = None,
                 unit: Optional[str] = None,
                 no_parse: Optional[bool] = False,
                 nonneg: Optional[bool] = False,
                 nonpos: Optional[bool] = False,
                 cplx: Optional[bool] = False,
                 imag: Optional[bool] = False,
                 symmetric: Optional[bool] = False,
                 diag: Optional[bool] = False,
                 hermitian: Optional[bool] = False,
                 boolean: Optional[bool] = False,
                 integer: Optional[bool] = False,
                 pos: Optional[bool] = False,
                 neg: Optional[bool] = False,
                 sparse: Optional[bool] = False,
                 ):
        OptzBase.__init__(self, name=name, info=info, unit=unit)
        self.no_parse = no_parse  # True to skip parsing the parameter
        self.sparse = sparse

        self.config = Config(name=self.class_name)  # `config` that can be exported

        self.config.add(OrderedDict((('nonneg', nonneg),
                                     ('nonpos', nonpos),
                                     ('complex', cplx),
                                     ('imag', imag),
                                     ('symmetric', symmetric),
                                     ('diag', diag),
                                     ('hermitian', hermitian),
                                     ('boolean', boolean),
                                     ('integer', integer),
                                     ('pos', pos),
                                     ('neg', neg),
                                     )))

    @ensure_symbols
    def parse(self):
        """
        Parse the parameter.

        Returns
        -------
        bool
            Returns True if the parsing is successful, False otherwise.
        """
        sub_map = self.om.rtn.syms.sub_map
        code_param = "param(**config)"
        for pattern, replacement, in sub_map.items():
            try:
                code_param = re.sub(pattern, replacement, code_param)
            except Exception as e:
                raise Exception(f"Error in parsing param <{self.name}>.\n{e}")
        self.code = code_param
        return True

    @ensure_mats_and_parsed
    def evaluate(self):
        """
        Evaluate the parameter.
        """
        if self.no_parse:
            return True

        config = self.config.as_dict()
        try:
            msg = f"Parameter <{self.name}> is set as sparse, "
            msg += "but the value is not sparse."
            if self.sparse:
                self.v = sps.csr_matrix(self.v)

            # Create the cvxpy.Parameter object
            self.optz = cp.Parameter(shape=self.v.shape, **config)
            self.optz.value = self.v
        except ValueError:
            msg = f"Parameter <{self.name}> has non-numeric value, "
            msg += "set `no_parse=True`."
            logger.warning(msg)
            self.no_parse = True
            return True
        except Exception as e:
            raise Exception(f"Error in evaluating param <{self.name}>.\n{e}")
        return True

    def update(self):
        """
        Update the Parameter value.
        """
        # NOTE: skip no_parse parameters
        if self.optz is None:
            return None
        self.optz.value = self.v
        return True

    def __repr__(self):
        return f'{self.__class__.__name__}: {self.name}'


class Var(OptzBase):
    """
    Base class for variables used in a routine.

    When `horizon` is provided, the variable will be expanded to a matrix,
    where rows are indexed by the source variable index and columns are
    indexed by the horizon index.

    Parameters
    ----------
    info : str, optional
        Descriptive information
    unit : str, optional
        Unit
    tex_name : str
        LaTeX-formatted variable symbol. Defaults to the value of ``name``.
    name : str, optional
        Variable name. One should typically assigning the name directly because
        it will be automatically assigned by the model. The value of ``name``
        will be the symbol name to be used in expressions.
    src : str, optional
        Source variable name. Defaults to the value of ``name``.
    model : str, optional
        Name of the owner model or group.
    horizon : ams.routines.RParam, optional
        Horizon idx.
    nonneg : bool, optional
        Non-negative variable
    nonpos : bool, optional
        Non-positive variable
    cplx : bool, optional
        Complex variable
    imag : bool, optional
        Imaginary variable
    symmetric : bool, optional
        Symmetric variable
    diag : bool, optional
        Diagonal variable
    psd : bool, optional
        Positive semi-definite variable
    nsd : bool, optional
        Negative semi-definite variable
    hermitian : bool, optional
        Hermitian variable
    boolean : bool, optional
        Boolean variable
    integer : bool, optional
        Integer variable
    pos : bool, optional
        Positive variable
    neg : bool, optional
        Negative variable

    Attributes
    ----------
    a : np.ndarray
        Variable address.
    _v : np.ndarray
        Local-storage of the variable value.
    rtn : ams.routines.Routine
        The owner routine instance.
    """

    def __init__(self,
                 name: Optional[str] = None,
                 tex_name: Optional[str] = None,
                 info: Optional[str] = None,
                 src: Optional[str] = None,
                 unit: Optional[str] = None,
                 model: Optional[str] = None,
                 shape: Optional[Union[tuple, int]] = None,
                 v0: Optional[str] = None,
                 horizon=None,
                 nonneg: Optional[bool] = False,
                 nonpos: Optional[bool] = False,
                 cplx: Optional[bool] = False,
                 imag: Optional[bool] = False,
                 symmetric: Optional[bool] = False,
                 diag: Optional[bool] = False,
                 psd: Optional[bool] = False,
                 nsd: Optional[bool] = False,
                 hermitian: Optional[bool] = False,
                 boolean: Optional[bool] = False,
                 integer: Optional[bool] = False,
                 pos: Optional[bool] = False,
                 neg: Optional[bool] = False,
                 ):
        self.name = name
        self.info = info
        self.unit = unit

        self.tex_name = tex_name if tex_name else name
        # instance of the owner Model
        self.owner = None
        # variable internal index inside a model (assigned in run time)
        self.id = None
        OptzBase.__init__(self, name=name, info=info, unit=unit)
        self.src = src
        self.is_group = False
        self.model = model  # indicate if this variable is a group variable
        self.owner = None  # instance of the owner model or group
        self.v0 = v0
        self.horizon = horizon
        self._shape = shape
        self._v = None
        self.a: np.ndarray = np.array([], dtype=int)

        self.config = Config(name=self.class_name)  # `config` that can be exported

        self.config.add(OrderedDict((('nonneg', nonneg),
                                     ('nonpos', nonpos),
                                     ('complex', cplx),
                                     ('imag', imag),
                                     ('symmetric', symmetric),
                                     ('diag', diag),
                                     ('psd', psd),
                                     ('nsd', nsd),
                                     ('hermitian', hermitian),
                                     ('boolean', boolean),
                                     ('integer', integer),
                                     ('pos', pos),
                                     ('neg', neg),
                                     )))

    @property
    def v(self):
        """
        Return the CVXPY variable value.
        """
        if self.optz is None:
            return None
        if self.optz.value is None:
            try:
                shape = self.optz.shape
                return np.zeros(shape)
            except AttributeError:
                return None
        else:
            return self.optz.value

    @v.setter
    def v(self, value):
        if self.optz is None:
            logger.info(f"Variable <{self.name}> is not initialized yet.")
        else:
            self.optz.value = value

    def get_idx(self):
        if self.is_group:
            return self.owner.get_idx()
        elif self.owner is None:
            logger.info(f'Variable <{self.name}> has no owner.')
            return None
        else:
            return self.owner.idx.v

    @ensure_symbols
    def parse(self):
        """
        Parse the variable.
        """
        sub_map = self.om.rtn.syms.sub_map
        # NOTE: number of rows is the size of the source variable
        if self.owner is not None:
            nr = self.owner.n
            nc = 0
            if self.horizon:
                # NOTE: numer of columns is the horizon if exists
                nc = int(self.horizon.n)
                shape = (nr, nc)
            else:
                shape = (nr,)
        elif isinstance(self._shape, int):
            shape = (self._shape,)
            nr = shape
            nc = 0
        elif isinstance(self._shape, tuple):
            shape = self._shape
            nr = shape[0]
            nc = shape[1] if len(shape) > 1 else 0
        else:
            raise ValueError(f"Invalid shape {self._shape}.")
        code_var = f"var({shape}, **config)"
        logger.debug(f" - Var <{self.name}>: {self.code}")
        for pattern, replacement, in sub_map.items():
            try:
                code_var = re.sub(pattern, replacement, code_var)
            except Exception as e:
                raise Exception(f"Error in parsing var <{self.name}>.\n{e}")
        self.code = code_var
        return True

    @ensure_mats_and_parsed
    def evaluate(self):
        """
        Evaluate the variable.

        Returns
        -------
        bool
            Returns True if the evaluation is successful, False otherwise.
        """
        # NOTE: in CVXPY, Config only allow lower case letters
        config = {}  # used in `self.code`
        for k, v in self.config.as_dict().items():
            if k == 'psd':
                config['PSD'] = v
            elif k == 'nsd':
                config['NSD'] = v
            elif k == 'bool':
                config['boolean'] = v
            else:
                config[k] = v
        msg = f" - Var <{self.name}>: {self.code}"
        logger.debug(pretty_long_message(msg, _prefix, max_length=_max_length))
        try:
            local_vars = {'self': self, 'config': config, 'cp': cp}
            self.optz = eval(self.code, {}, local_vars)
        except Exception as e:
            raise Exception(f"Error in evaluating var <{self.name}>.\n{e}")
        return True

    def __repr__(self):
        return f'{self.__class__.__name__}: {self.owner.__class__.__name__}.{self.name}'


class Constraint(OptzBase):
    """
    Base class for constraints.

    This class is used as a template for defining constraints. Each
    instance of this class represents a single constraint.

    Parameters
    ----------
    name : str, optional
        A user-defined name for the constraint.
    e_str : str, optional
        A mathematical expression representing the constraint.
    info : str, optional
        Additional informational text about the constraint.
    is_eq : str, optional
        Flag indicating if the constraint is an equality constraint. False indicates
        an inequality constraint in the form of `<= 0`.

    Attributes
    ----------
    is_disabled : bool
        Flag indicating if the constraint is disabled, False by default.
    rtn : ams.routines.Routine
        The owner routine instance.
    is_disabled : bool, optional
        Flag indicating if the constraint is disabled, False by default.
    dual : float, optional
        The dual value of the constraint.
    code : str, optional
        The code string for the constraint
    """

    def __init__(self,
                 name: Optional[str] = None,
                 e_str: Optional[str] = None,
                 info: Optional[str] = None,
                 is_eq: Optional[bool] = False,
                 ):
        OptzBase.__init__(self, name=name, info=info)
        self.e_str = e_str
        self.is_eq = is_eq
        self.is_disabled = False
        self.dual = None
        self.code = None

    @ensure_symbols
    def parse(self):
        """
        Parse the constraint.
        """
        # parse the expression str
        sub_map = self.om.rtn.syms.sub_map
        code_constr = self.e_str
        for pattern, replacement in sub_map.items():
            try:
                code_constr = re.sub(pattern, replacement, code_constr)
            except TypeError as e:
                raise TypeError(f"Error in parsing constr <{self.name}>.\n{e}")
        # parse the constraint type
        code_constr += " == 0" if self.is_eq else " <= 0"
        # store the parsed expression str code
        self.code = code_constr
        msg = f" - Constr <{self.name}>: {self.e_str}"
        logger.debug(pretty_long_message(msg, _prefix, max_length=_max_length))
        return True

    def _evaluate_expression(self, code, local_vars=None):
        """
        Helper method to evaluate the expression code.

        Parameters
        ----------
        code : str
            The code string representing the expression.

        Returns
        -------
        cp.Expression
            The evaluated cvxpy expression.
        """
        return eval(code, {}, local_vars)

    @ensure_mats_and_parsed
    def evaluate(self):
        """
        Evaluate the constraint.
        """
        msg = f" - Constr <{self.name}>: {self.code}"
        logger.debug(pretty_long_message(msg, _prefix, max_length=_max_length))
        try:
            local_vars = {'self': self, 'cp': cp, 'sub_map': self.om.rtn.syms.val_map}
            self.optz = self._evaluate_expression(self.code, local_vars=local_vars)
        except Exception as e:
            raise Exception(f"Error in evaluating constr <{self.name}>.\n{e}")

    def __repr__(self):
        enabled = 'OFF' if self.is_disabled else 'ON'
        out = f"{self.class_name}: {self.name} [{enabled}]"
        return out

    @property
    def e(self):
        """
        Return the calculated constraint LHS value.
        Note that `v` should be used primarily as it is obtained
        from the solver directly.

        `e` is for debugging purpose. For a successfully solved problem,
        `e` should equal to `v`. However, when a problem is infeasible
        or unbounded, `e` can be used to check the constraint LHS value.
        """
        if self.code is None:
            logger.info(f"Constraint <{self.name}> is not parsed yet.")
            return None

        val_map = self.om.rtn.syms.val_map
        code = self.code
        for pattern, replacement in val_map.items():
            try:
                code = re.sub(pattern, replacement, code)
            except TypeError as e:
                raise TypeError(e)

        try:
            logger.debug(pretty_long_message(f"Value code: {code}",
                                             _prefix, max_length=_max_length))
            local_vars = {'self': self, 'np': np, 'cp': cp, 'val_map': val_map}
            return self._evaluate_expression(code, local_vars)
        except Exception as e:
            logger.error(f"Error in calculating constr <{self.name}>.\n{e}")
            return None

    @property
    def v(self):
        """
        Return the CVXPY constraint LHS value.
        """
        if self.optz is None:
            return None
        if self.optz._expr.value is None:
            try:
                shape = self._expr.shape
                return np.zeros(shape)
            except AttributeError:
                return None
        else:
            return self.optz._expr.value

    @v.setter
    def v(self, value):
        raise AttributeError("Cannot set the value of the constraint.")


class Objective(OptzBase):
    """
    Base class for objective functions.

    This class serves as a template for defining objective functions. Each
    instance of this class represents a single objective function that can
    be minimized or maximized depending on the sense ('min' or 'max').

    Parameters
    ----------
    name : str, optional
        A user-defined name for the objective function.
    e_str : str, optional
        A mathematical expression representing the objective function.
    info : str, optional
        Additional informational text about the objective function.
    sense : str, optional
        The sense of the objective function, default to 'min'.
        `min` for minimization and `max` for maximization.

    Attributes
    ----------
    v : NoneType
        The value of the objective function. It needs to be set through
        computation.
    rtn : ams.routines.Routine
        The owner routine instance.
    code : str
        The code string for the objective function.
    """

    def __init__(self,
                 name: Optional[str] = None,
                 e_str: Optional[str] = None,
                 info: Optional[str] = None,
                 unit: Optional[str] = None,
                 sense: Optional[str] = 'min'):
        OptzBase.__init__(self, name=name, info=info, unit=unit)
        self.e_str = e_str
        self.sense = sense
        self.code = None

    @property
    def e(self):
        """
        Return the calculated objective value.

        Note that `v` should be used primarily as it is obtained
        from the solver directly.

        `e` is for debugging purpose. For a successfully solved problem,
        `e` should equal to `v`. However, when a problem is infeasible
        or unbounded, `e` can be used to check the objective value.
        """
        if self.code is None:
            logger.info(f"Objective <{self.name}> is not parsed yet.")
            return None

        val_map = self.om.rtn.syms.val_map
        code = self.code
        for pattern, replacement in val_map.items():
            try:
                code = re.sub(pattern, replacement, code)
            except TypeError as e:
                logger.error(f"Error in parsing value for obj <{self.name}>.")
                raise e

        try:
            logger.debug(pretty_long_message(f"Value code: {code}",
                                             _prefix, max_length=_max_length))
            local_vars = {'self': self, 'np': np, 'cp': cp, 'val_map': val_map}
            return self._evaluate_expression(code, local_vars)
        except Exception as e:
            logger.error(f"Error in calculating obj <{self.name}>.\n{e}")
            return None

    @property
    def v(self):
        """
        Return the CVXPY objective value.
        """
        if self.optz is None:
            return None
        else:
            return self.optz.value

    @v.setter
    def v(self, value):
        raise AttributeError("Cannot set the value of the objective function.")

    @ensure_symbols
    def parse(self):
        """
        Parse the objective function.

        Returns
        -------
        bool
            Returns True if the parsing is successful, False otherwise.
        """
        # parse the expression str
        sub_map = self.om.rtn.syms.sub_map
        code_obj = self.e_str
        for pattern, replacement, in sub_map.items():
            try:
                code_obj = re.sub(pattern, replacement, code_obj)
            except Exception as e:
                raise Exception(f"Error in parsing obj <{self.name}>.\n{e}")
        # store the parsed expression str code
        self.code = code_obj
        if self.sense not in ['min', 'max']:
            raise ValueError(f'Objective sense {self.sense} is not supported.')
        sense = 'cp.Minimize' if self.sense == 'min' else 'cp.Maximize'
        self.code = f"{sense}({code_obj})"
        msg = f" - Objective <{self.name}>: {self.code}"
        logger.debug(pretty_long_message(msg, _prefix, max_length=_max_length))
        return True

    @ensure_mats_and_parsed
    def evaluate(self):
        """
        Evaluate the objective function.

        Returns
        -------
        bool
            Returns True if the evaluation is successful, False otherwise.
        """
        logger.debug(f" - Objective <{self.name}>: {self.e_str}")
        try:
            local_vars = {'self': self, 'cp': cp}
            self.optz = self._evaluate_expression(self.code, local_vars=local_vars)
        except Exception as e:
            raise Exception(f"Error in evaluating obj <{self.name}>.\n{e}")
        return True

    def _evaluate_expression(self, code, local_vars=None):
        """
        Helper method to evaluate the expression code.

        Parameters
        ----------
        code : str
            The code string representing the expression.

        Returns
        -------
        cp.Expression
            The evaluated cvxpy expression.
        """
        return eval(code, {}, local_vars)

    def __repr__(self):
        return f"{self.class_name}: {self.name} [{self.sense.upper()}]"


class OModel:
    """
    Base class for optimization models.

    Parameters
    ----------
    routine: Routine
        Routine that to be modeled.

    Attributes
    ----------
    prob: cvxpy.Problem
        Optimization model.
    params: OrderedDict
        Parameters.
    vars: OrderedDict
        Decision variables.
    constrs: OrderedDict
        Constraints.
    obj: Objective
        Objective function.
    initialized: bool
        Flag indicating if the model is initialized.
    parsed: bool
        Flag indicating if the model is parsed.
    evaluated: bool
        Flag indicating if the model is evaluated.
    """

    def __init__(self, routine):
        self.rtn = routine
        self.prob = None
        self.params = OrderedDict()
        self.vars = OrderedDict()
        self.constrs = OrderedDict()
        self.obj = None
        self.initialized = False
        self.parsed = False
        self.evaluated = False
        self.finalized = False

    @ensure_symbols
    def parse(self, force=False):
        """
        Parse the optimization model from the symbolic description.

        This method should be called after the routine symbols are generated
        `self.rtn.syms.generate_symbols()`. It parses the following components
        of the optimization model: parameters, decision variables, constraints,
        objective function, and expressions.

        Parameters
        ----------
        force : bool, optional
            Flag indicating if to force the parsing.

        Returns
        -------
        bool
            Returns True if the parsing is successful, False otherwise.
        """
        if self.parsed and not force:
            logger.debug("Model is already parsed.")
            return self.parsed
        t, _ = elapsed()
        # --- add RParams and Services as parameters ---
        logger.warning(f'Parsing OModel for <{self.rtn.class_name}>')
        for key, val in self.rtn.params.items():
            if not val.no_parse:
                val.parse()

        # --- add decision variables ---
        for key, val in self.rtn.vars.items():
            val.parse()

        # --- add constraints ---
        for key, val in self.rtn.constrs.items():
            val.parse()

        # --- parse objective functions ---
        if self.rtn.type != 'PF':
            if self.rtn.obj is not None:
                try:
                    self.rtn.obj.parse()
                except Exception as e:
                    raise Exception(f"Failed to parse Objective <{self.rtn.obj.name}>.\n{e}")
            else:
                logger.warning(f"{self.rtn.class_name} has no objective function!")
                self.parsed = False
                return self.parsed

        # --- parse expressions ---
        for key, val in self.rtn.exprs.items():
            try:
                val.parse()
            except Exception as e:
                raise Exception(f"Failed to parse ExpressionCalc <{key}>.\n{e}")
        _, s = elapsed(t)
        logger.debug(f"  -> Parsed in {s}")

        self.parsed = True
        return self.parsed

    def _evaluate_params(self):
        """
        Evaluate the parameters.
        """
        for key, val in self.rtn.params.items():
            try:
                val.evaluate()
                setattr(self, key, val.optz)
            except Exception as e:
                raise Exception(f"Failed to evaluate Param <{key}>.\n{e}")

    def _evaluate_vars(self):
        """
        Evaluate the decision variables.
        """
        for key, val in self.rtn.vars.items():
            try:
                val.evaluate()
                setattr(self, key, val.optz)
            except Exception as e:
                raise Exception(f"Failed to evaluate Var <{key}>.\n{e}")

    def _evaluate_constrs(self):
        """
        Evaluate the constraints.
        """
        for key, val in self.rtn.constrs.items():
            try:
                val.evaluate()
                setattr(self, key, val.optz)
            except Exception as e:
                raise Exception(f"Failed to evaluate Constr <{key}>.\n{e}")

    def _evaluate_obj(self):
        """
        Evaluate the objective function.
        """
        # NOTE: since we already have the attribute `obj`,
        # we can update it rather than setting it
        if self.rtn.type != 'PF':
            self.rtn.obj.evaluate()
            self.obj = self.rtn.obj.optz

    def _evaluate_exprs(self):
        """
        Evaluate the expressions.
        """
        for key, val in self.rtn.exprs.items():
            try:
                val.evaluate()
            except Exception as e:
                raise Exception(f"Failed to evaluate ExpressionCalc <{key}>.\n{e}")

    @ensure_mats_and_parsed
    def evaluate(self, force=False):
        """
        Evaluate the optimization model.

        This method should be called after `self.parse()`. It evaluates the following
        components of the optimization model: parameters, decision variables, constraints,
        objective function, and expressions.

        Parameters
        ----------
        force : bool, optional
            Flag indicating if to force the evaluation

        Returns
        -------
        bool
            Returns True if the evaluation is successful, False otherwise.
        """
        if self.evaluated and not force:
            logger.debug("Model is already evaluated.")
            return self.evaluated
        logger.warning(f"Evaluating OModel for <{self.rtn.class_name}>")
        t, _ = elapsed()

        self._evaluate_params()
        self._evaluate_vars()
        self._evaluate_constrs()
        self._evaluate_obj()
        self._evaluate_exprs()

        self.evaluated = True
        _, s = elapsed(t)
        logger.debug(f" -> Evaluated in {s}")
        return self.evaluated

    def finalize(self, force=False):
        """
        Finalize the optimization model.

        This method should be called after `self.evaluate()`. It assemble the optimization
        problem from the evaluated components.

        Returns
        -------
        bool
            Returns True if the finalization is successful, False otherwise.
        """
        # NOTE: for power flow type, we skip the finalization
        if self.rtn.type == 'PF':
            self.finalized = True
            return self.finalized
        if self.finalized and not force:
            logger.debug("Model is already finalized.")
            return self.finalized
        logger.warning(f"Finalizing OModel for <{self.rtn.class_name}>")
        t, _ = elapsed()

        # Collect constraints that are not disabled
        constrs_add = [val.optz for key, val in self.rtn.constrs.items(
        ) if not val.is_disabled and val is not None]
        # Construct the problem using cvxpy.Problem
        self.prob = cp.Problem(self.obj, constrs_add)

        _, s = elapsed(t)
        logger.debug(f" -> Finalized in {s}")
        self.finalized = True
        return self.finalized

    def init(self, force=False):
        """
        Set up the optimization model from the symbolic description.

        This method initializes the optimization model by parsing decision variables,
        constraints, and the objective function from the associated routine.

        Parameters
        ----------
        force : bool, optional
            Flag indicating if to force the OModel initialization.
            If True, following methods will be called by force: `self.parse()`,
            `self.evaluate()`, `self.finalize()`

        Returns
        -------
        bool
            Returns True if the setup is successful, False otherwise.
        """
        if self.initialized and not force:
            logger.debug("OModel is already initialized.")
            return self.initialized

        t, _ = elapsed()

        self.parse(force=force)
        self.evaluate(force=force)
        self.finalize(force=force)

        _, s = elapsed(t)
        self.initialized = True
        logger.debug(f"OModel for <{self.rtn.class_name}> initialized in {s}")

        return self.initialized

    @property
    def class_name(self):
        """
        Return the class name
        """
        return self.__class__.__name__

    def _register_attribute(self, key, value):
        """
        Register a pair of attributes to OModel instance.

        Called within ``__setattr__``, this is where the magic happens.
        Subclass attributes are automatically registered based on the variable type.
        """
        if isinstance(value, cp.Variable):
            self.vars[key] = value
        elif isinstance(value, cp.Constraint):
            self.constrs[key] = value
        elif isinstance(value, cp.Parameter):
            self.params[key] = value

    def __setattr__(self, name: str, value: Any):
        super().__setattr__(name, value)
        self._register_attribute(name, value)

    def update(self, params):
        """
        Update the Parameter values.

        Parameters
        ----------
        params: list
            List of parameters to be updated.
        """
        for param in params:
            param.update()
        return True

    def __repr__(self) -> str:
        return f'{self.rtn.class_name}.{self.__class__.__name__} at {hex(id(self))}'



"""
Base class for parameters.
"""


import logging

from typing import Optional, Iterable

import numpy as np
from scipy.sparse import issparse

from ams.opt.omodel import Param

logger = logging.getLogger(__name__)


class RParam(Param):
    """
    Class for parameters used in a routine.
    This class is developed to simplify the routine definition.

    `RParm` is further used to define `Parameter` in the optimization model.

    `no_parse` is used to skip parsing the `RParam` in optimization
    model.
    It means that the `RParam` will not be added to the optimization model.
    This is useful when the RParam contains non-numeric values,
    or it is not necessary to be added to the optimization model.

    Parameters
    ----------
    name : str, optional
        Name of this parameter. If not provided, `name` will be set
        to the attribute name.
    tex_name : str, optional
        LaTeX-formatted parameter name. If not provided, `tex_name`
        will be assigned the same as `name`.
    info : str, optional
        A description of this parameter
    src : str, optional
        Source name of the parameter.
    unit : str, optional
        Unit of the parameter.
    model : str, optional
        Name of the owner model or group.
    v : np.ndarray, optional
        External value of the parameter.
    indexer : str, optional
        Indexer of the parameter.
    imodel : str, optional
        Name of the owner model or group of the indexer.
    no_parse: bool, optional
        True to skip parsing the parameter.
    nonneg: bool, optional
        True to set the parameter as non-negative.
    nonpos: bool, optional
        True to set the parameter as non-positive.
    cplx: bool, optional
        True to set the parameter as complex.
    imag: bool, optional
        True to set the parameter as imaginary.
    symmetric: bool, optional
        True to set the parameter as symmetric.
    diag: bool, optional
        True to set the parameter as diagonal.
    hermitian: bool, optional
        True to set the parameter as hermitian.
    boolean: bool, optional
        True to set the parameter as boolean.
    integer: bool, optional
        True to set the parameter as integer.
    pos: bool, optional
        True to set the parameter as positive.
    neg: bool, optional
        True to set the parameter as negative.
    sparse: bool, optional
        True to set the parameter as sparse.

    Examples
    --------
    Example 1: Define a routine parameter from a source model or group.

    In this example, we define the parameter `cru` from the source model
    `SFRCost` with the parameter `cru`.

    >>> self.cru = RParam(info='RegUp reserve coefficient',
    >>>                   tex_name=r'c_{r,u}',
    >>>                   unit=r'$/(p.u.)',
    >>>                   name='cru',
    >>>                   src='cru',
    >>>                   model='SFRCost'
    >>>                   )

    Example 2: Define a routine parameter with a user-defined value.

    In this example, we define the parameter with a user-defined value.
    TODO: Add example
    """

    def __init__(self,
                 name: Optional[str] = None,
                 tex_name: Optional[str] = None,
                 info: Optional[str] = None,
                 src: Optional[str] = None,
                 unit: Optional[str] = None,
                 model: Optional[str] = None,
                 v: Optional[np.ndarray] = None,
                 indexer: Optional[str] = None,
                 imodel: Optional[str] = None,
                 expand_dims: Optional[int] = None,
                 no_parse: Optional[bool] = False,
                 nonneg: Optional[bool] = False,
                 nonpos: Optional[bool] = False,
                 cplx: Optional[bool] = False,
                 imag: Optional[bool] = False,
                 symmetric: Optional[bool] = False,
                 diag: Optional[bool] = False,
                 hermitian: Optional[bool] = False,
                 boolean: Optional[bool] = False,
                 integer: Optional[bool] = False,
                 pos: Optional[bool] = False,
                 neg: Optional[bool] = False,
                 sparse: Optional[list] = None,
                 ):
        Param.__init__(self, nonneg=nonneg, nonpos=nonpos,
                       cplx=cplx, imag=imag, symmetric=symmetric,
                       diag=diag, hermitian=hermitian, boolean=boolean,
                       integer=integer, pos=pos, neg=neg, sparse=sparse)
        self.name = name
        self.tex_name = tex_name if (tex_name is not None) else name
        self.info = info
        self.src = name if (src is None) else src
        self.unit = unit
        self.is_group = False
        self.model = model  # name of a group or model
        self.indexer = indexer  # name of the indexer
        self.imodel = imodel  # name of a group or model of the indexer
        self.expand_dims = expand_dims
        self.no_parse = no_parse
        self.owner = None  # instance of the owner model or group
        self.is_ext = False  # indicate if the value is set externally
        self._v = None  # external value
        if v is not None:
            self._v = v
            self.is_ext = True

    # FIXME: might need a better organization
    @property
    def v(self):
        """
        The value of the parameter.

        Notes
        -----
        - This property is a wrapper for the ``get`` method of the owner class.
        - The value will sort by the indexer if indexed, used for optmization modeling.
        """
        out = None
        if self.sparse and self.expand_dims is not None:
            msg = 'Sparse matrix does not support expand_dims.'
            raise NotImplementedError(msg)
        if self.indexer is None:
            if self.is_ext:
                if issparse(self._v):
                    out = self._v.toarray()
                else:
                    out = self._v
            elif self.is_group:
                out = self.owner.get(src=self.src, attr='v',
                                     idx=self.owner.get_idx())
            else:
                src_param = getattr(self.owner, self.src)
                out = getattr(src_param, 'v')
        else:
            try:
                imodel = getattr(self.rtn.system, self.imodel)
            except AttributeError:
                msg = f'Indexer source model <{self.imodel}> not found, '
                msg += 'likely a modeling error.'
                raise AttributeError(msg)
            try:
                sorted_idx = self.owner.find_idx(keys=self.indexer, values=imodel.get_idx())
            except AttributeError:
                sorted_idx = self.owner.idx.v
            except Exception as e:
                raise e
            model = getattr(self.rtn.system, self.model)
            out = model.get(src=self.src, attr='v', idx=sorted_idx)
        if self.expand_dims is not None:
            out = np.expand_dims(out, axis=self.expand_dims)
        return out

    @property
    def shape(self):
        """
        Return the shape of the parameter.
        """
        if self.is_ext:
            return np.shape(self._v)
        elif self.no_parse:
            return None
        else:
            return np.shape(self.v)

    @property
    def dtype(self):
        """
        Return the data type of the parameter value.
        """
        if isinstance(self.v, (str, bytes)):
            return str
        elif isinstance(self.v, Iterable):
            return type(self.v[0])
        else:
            return type(self.v)

    @property
    def n(self):
        """
        Return the szie of the parameter.
        """
        if self.is_ext:
            return self._v.shape[0]
        else:
            return self.owner.n

    @property
    def class_name(self):
        """
        Return the class name
        """
        return self.__class__.__name__

    def __repr__(self):
        owner = self.owner.__class__.__name__ if self.owner is not None else self.rtn.__class__.__name__
        postfix = '' if self.src is None else f'.{self.src}'
        return f'{self.__class__.__name__}: {owner}' + postfix

    def get_idx(self):
        """
        Get the index of the parameter.

        Returns
        -------
        idx : list
            Index of the parameter.

        Notes
        -----
        - The value will sort by the indexer if indexed.
        """
        if self.indexer is None:
            if self.is_group:
                return self.owner.get_idx()
            elif self.owner is None:
                logger.info(f'Param <{self.name}> has no owner.')
                return None
            elif hasattr(self.owner, 'idx'):
                return self.owner.idx.v
            else:
                logger.info(f'Param <{self.name}> owner <{self.owner.class_name}> has no idx.')
                return None
        else:
            try:
                imodel = getattr(self.rtn.system, self.imodel)
            except AttributeError:
                msg = f'Indexer source model <{self.imodel}> not found, '
                msg += 'likely a modeling error.'
                raise AttributeError(msg)
            try:
                sorted_idx = self.owner.find_idx(keys=self.indexer, values=imodel.get_idx())
            except AttributeError:
                msg = f'Indexer <{self.indexer}> not found in <{self.imodel}>, '
                msg += 'likely a modeling error.'
                raise AttributeError(msg)
            return sorted_idx



"""
Test `Jumper` class.
"""

import unittest

import ams


class TessJumper(unittest.TestCase):
    """
    Test `Jumper` class.
    """

    def setUp(self) -> None:
        """
        Test setup.
        """
        self.sp = ams.load(ams.get_case("5bus/pjm5bus_jumper.xlsx"),
                           setup=True, no_output=True, default_config=True)

    def test_to_andes(self):
        """
        Test `to_andes` method when `Jumper` exists.
        """
        sa = self.sp.to_andes()
        self.assertEqual(sa.Jumper.n, 1)
